"""Globi - Data and simulations for the GloBI project.

A Python package for data science and geospatial analysis.
"""

__version__ = "0.1.0"

# Import main classes and functions (lazy imports to avoid dependency issues)
# Users can import directly from submodules if needed:
# from globi.globi_builder import GloBIBuildingSpec, run_simulation
# from globi.globi_builder import post_process_results, save_results_json
# from globi.experiment_config import GloBISpecInputScythe, simulate_globi_building

__all__ = [
    "__version__",
]



"""Type literals for the GloBI project."""

from typing import Literal

BasementAtticOccupationConditioningStatus = Literal[
    "none",
    "unoccupied_unconditioned",
    "unoccupied_conditioned",
    "occupied_unconditioned",
    "occupied_conditioned",
]

OccupiedOptions: list[BasementAtticOccupationConditioningStatus] = [
    "occupied_unconditioned",
    "occupied_conditioned",
]

UnoccupiedOptions: list[BasementAtticOccupationConditioningStatus] = [
    "none",
    "unoccupied_unconditioned",
    "unoccupied_conditioned",
]

ConditionedOptions: list[BasementAtticOccupationConditioningStatus] = [
    "occupied_conditioned",
    "unoccupied_conditioned",
]

UnconditionedOptions: list[BasementAtticOccupationConditioningStatus] = [
    "none",
    "unoccupied_unconditioned",
    "occupied_unconditioned",
]



"""Worker main script."""

from scythe.worker import ScytheWorkerConfig

from globi.pipelines import *  # noqa: F403

conf = ScytheWorkerConfig()


def main():
    """Main function for the worker."""
    conf.start()


if __name__ == "__main__":
    conf.start()



"""Errors for GIS preprocessor."""


class GISPreprocessorError(Exception):
    """Exception for GIS preprocessor errors."""


class GISFileHasNoFeaturesError(GISPreprocessorError):
    """Exception for when a GIS file has no features."""

    def __init__(self):
        """Initialize the GIS file has no features error."""
        super().__init__("GIS data contains no features.")


class GISFileHasInvalidCategoricalSemanticFieldError(GISPreprocessorError):
    """Exception for when a GIS file has invalid categorical semantic fields."""

    def __init__(self, field_name: str, invalid_values: list[str], options: list[str]):
        """Initialize the GIS file has invalid semantic field error."""
        self.field_name = field_name
        self.invalid_values = invalid_values
        self.options = options
        printable_invalid_values = (
            [*invalid_values[:5], "..."] if len(invalid_values) > 5 else invalid_values
        )
        super().__init__(
            f"Field '{field_name}' has values which are not in the allowed options: "
            f"{printable_invalid_values}. Allowed options: {options}"
        )


class GISFileHasOutOfBoundsNumericSemanticFieldError(GISPreprocessorError):
    """Exception for when a GIS file has out of bounds numeric semantic fields."""

    def __init__(
        self, field_name: str, values: list[float], min_val: float, max_val: float
    ):
        """Initialize the GIS file has out of bounds numeric semantic field error."""
        self.field_name = field_name
        self.values = values
        self.min_val = min_val
        self.max_val = max_val
        printable_values = [*values[:5], "..."] if len(values) > 5 else values
        super().__init__(
            f"Field '{field_name}' has values which are out of bounds: {printable_values}. "
            f"Allowed values: {min_val} to {max_val}"
        )


class GISFileHasSemanticFieldWithNoValidatorError(GISPreprocessorError):
    """Exception for when a GIS file has a semantic field with no validator."""

    def __init__(self, field_name: str):
        """Initialize the GIS file has semantic field with no validator error."""
        self.field_name = field_name
        super().__init__(f"Field '{field_name}' has no validator.")


class GISFileHasMissingColumnsError(GISPreprocessorError):
    """Exception for when a GIS file has missing columns."""

    def __init__(self, missing_cols: list[str], available_cols: list[str]):
        """Initialize the GIS file has missing columns error."""
        self.missing_cols = missing_cols
        self.available_cols = available_cols
        super().__init__(
            f"GIS file has missing columns: {missing_cols}. Available columns: {available_cols}"
        )


class GISFileHasNoCRSError(GISPreprocessorError):
    """Exception for when a GIS file has no CRS."""

    def __init__(self):
        """Initialize the GIS file has no CRS error."""
        super().__init__(
            "GIS file has no crs.  Please set the CRS before running this script."
        )


class GISFileHasUnexpectedCRSError(GISPreprocessorError):
    """Exception for when a GIS file has an unexpected CRS."""

    def __init__(self, expected_crs: str, actual_crs: str, allow_mercator: bool = True):
        """Initialize the GIS file has unexpected CRS error."""
        self.expected_crs = expected_crs
        self.actual_crs = actual_crs
        self.allow_mercator = allow_mercator
        printable_crs = {"EPSG:4326"}
        if allow_mercator:
            printable_crs.add("EPSG:3857")
        printable_crs.add(expected_crs)
        super().__init__(
            f"GIS file has unexpected CRS: {actual_crs}. Expected CRS: {list(printable_crs)}.  Please update the CRS to one of the expected values before running this script."
        )


class GISFileMissingBothHeightAndFloorsError(GISPreprocessorError):
    """Exception for when a GIS file is missing both height and floors."""

    def __init__(self):
        """Initialize the GIS file missing both height and floors error."""
        super().__init__("GIS file is missing both height and floors.")


class GISFileHasNonNumericHeightError(GISPreprocessorError):
    """Exception for when a GIS file has a non-numeric height."""

    def __init__(self, height_col: str):
        """Initialize the GIS file has non-numeric height error."""
        self.height_col = height_col
        super().__init__(f"GIS file has a non-numeric height column: {height_col}.")


class GISFileHasNonNumericFloorsError(GISPreprocessorError):
    """Exception for when a GIS file has a non-numeric floors column."""

    def __init__(self, nfloors_col: str):
        """Initialize the GIS file has non-numeric floors error."""
        self.nfloors_col = nfloors_col
        super().__init__(f"GIS file has a non-numeric floors column: {nfloors_col}.")


class SemanticFieldsFileHasNoBuildingIDColumnError(GISPreprocessorError):
    """Exception for when a semantic fields file has no building ID column."""

    def __init__(self):
        """Initialize the semantic fields file has no building ID column error."""
        super().__init__("Semantic fields file is missing a building ID column.")


class GISFileHasNoBuildingIDColumnError(GISPreprocessorError):
    """Exception for when a GIS file has no building ID column."""

    def __init__(self):
        """Initialize the GIS file has no building ID column error."""
        super().__init__("GIS file is missing a building ID column.")


class GISFileHasMissingBuildingIDsError(GISPreprocessorError):
    """Exception for when a GIS file has missing building IDs."""

    def __init__(self):
        """Initialize the GIS file has missing building IDs error."""
        super().__init__("GIS file is missing building ID in some rows.")


class GISFileHasNonUniqueBuildingIDsError(GISPreprocessorError):
    """Exception for when a GIS file has non-unique building IDs."""

    def __init__(self):
        """Initialize the GIS file has non-unique building IDs error."""
        super().__init__("GIS file has non-unique building IDs.")



"""Geometry utilities."""

import logging
from collections.abc import Callable
from typing import cast

import geopandas as gpd
import numpy as np
import pandas as pd
from pydantic import BaseModel
from scipy.spatial import cKDTree
from shapely.geometry import Point, Polygon

logger = logging.getLogger(__name__)


class InjectedGeometryColumnMap(BaseModel):
    """A map of the columns that are injected into the GeoDataFrame."""

    Geometry_Cart_col: str
    Footprint_Area_col: str
    Rotated_Rectangle_col: str
    Rotated_Rectangle_Area_Ratio_col: str
    Short_Edge_col: str
    Long_Edge_col: str
    Long_Edge_Angle_col: str
    Aspect_Ratio_col: str

    @property
    def all_cols(self) -> list[str]:
        """Get all the column names as a list."""
        return list(self.model_dump().values())


def inject_rotated_rectangles(
    gdf: gpd.GeoDataFrame, cart_crs: str
) -> tuple[gpd.GeoDataFrame, InjectedGeometryColumnMap]:
    """Inject rotated rectangles into a GeoDataFrame.

    This will fit a rectangle around the geometry of each building, and additionally compute
    some useful summary features.

    Adds the following columns:
        - geometry_cart: the geometry in the cartesian CRS
        - footprint_area: the area of the footprint of the building
        - rotated_rectangle: the minimum rotated rectangle that bounds the building
        - rotated_rectangle_area_ratio: the ratio of the area of the rotated rectangle to the footprint area
        - short_edge: the shorter edge of the rotated rectangle
        - long_edge: the longer edge of the rotated rectangle
        - long_edge_angle: the angle of the long edge of the rotated rectangle
        - aspect_ratio: the ratio of the long edge to the short edge of the rotated rectangle

    Args:
        gdf (gpd.GeoDataFrame): The GeoDataFrame of buildings.
        cart_crs (str): The cartesian CRS to project the buildings to.

    Returns:
        gdf (gpd.GeoDataFrame): The GeoDataFrame with the rotated rectangles.
        injected_cols (list[str]): The injected columns
    """
    injected_column_map = InjectedGeometryColumnMap(
        Geometry_Cart_col="GLOBI_GEOMETRY_CART",
        Footprint_Area_col="GLOBI_FOOTPRINT_AREA",
        Rotated_Rectangle_col="GLOBI_ROTATED_RECTANGLE",
        Rotated_Rectangle_Area_Ratio_col="GLOBI_ROTATED_RECTANGLE_AREA_RATIO",
        Short_Edge_col="GLOBI_SHORT_EDGE",
        Long_Edge_col="GLOBI_LONG_EDGE",
        Long_Edge_Angle_col="GLOBI_LONG_EDGE_ANGLE",
        Aspect_Ratio_col="GLOBI_ASPECT_RATIO",
    )

    gdf_cart = cast(gpd.GeoDataFrame, gdf.to_crs(cart_crs))
    gdf[injected_column_map.Geometry_Cart_col] = gdf_cart.geometry
    gdf[injected_column_map.Footprint_Area_col] = gdf_cart.geometry.area
    gdf[injected_column_map.Rotated_Rectangle_col] = (
        gdf_cart.geometry.minimum_rotated_rectangle()
    )
    gdf[injected_column_map.Rotated_Rectangle_Area_Ratio_col] = (
        gdf[injected_column_map.Rotated_Rectangle_col].area
        / gdf[injected_column_map.Footprint_Area_col]
    )
    a1 = cast(
        pd.DataFrame,
        gdf[injected_column_map.Rotated_Rectangle_col].apply(
            lambda x: pd.Series({
                "x": x.exterior.coords[0][0],
                "y": x.exterior.coords[0][1],
            })
        ),
    )
    b1 = cast(
        pd.DataFrame,
        gdf[injected_column_map.Rotated_Rectangle_col].apply(
            lambda x: pd.Series({
                "x": x.exterior.coords[1][0],
                "y": x.exterior.coords[1][1],
            })
        ),
    )
    a2 = cast(
        pd.DataFrame,
        gdf[injected_column_map.Rotated_Rectangle_col].apply(
            lambda x: pd.Series({
                "x": x.exterior.coords[2][0],
                "y": x.exterior.coords[2][1],
            })
        ),
    )
    d1 = ((a1 - b1) ** 2).sum(axis=1) ** 0.5
    d2 = ((a2 - b1) ** 2).sum(axis=1) ** 0.5
    d1_is_long = d1 > d2
    short = d1.where(d1 < d2, d2)
    long = d1.where(d1 > d2, d2)
    long_edge_vector = (b1 - a1) * d1_is_long.values[:, None] + (
        a2 - b1
    ) * ~d1_is_long.values[:, None]
    long_edge_angle = np.arctan2(long_edge_vector.y, long_edge_vector.x)
    aspect_ratio = long / short
    gdf[injected_column_map.Short_Edge_col] = short
    gdf[injected_column_map.Long_Edge_col] = long
    gdf[injected_column_map.Long_Edge_Angle_col] = long_edge_angle
    gdf[injected_column_map.Aspect_Ratio_col] = aspect_ratio

    return (gdf, injected_column_map)


class InjectedNeighborColumnMap(BaseModel):
    """A map of the columns that are injected into the GeoDataFrame."""

    Neighbor_Ixs_col: str
    Neighbor_Polys_col: str
    Neighbor_Heights_col: str
    Neighbor_Floors_col: str


def inject_neighbor_ixs(
    gdf: gpd.GeoDataFrame,
    injected_geometry_col_map: InjectedGeometryColumnMap,
    neighbor_threshold: float = 50,
    remove_intersections: bool = True,
    log_fn: Callable | None = None,
) -> tuple[gpd.GeoDataFrame, InjectedNeighborColumnMap]:
    """Compute neighbors for each building in a GeoDataFrame.

    Note that this does mutate the input GeoDataFrame (in addition to returning it).

    Args:
        gdf (gpd.GeoDataFrame): The GeoDataFrame of buildings.
        injected_geometry_col_map (InjectedGeometryColumnMap): The injected geometry columns.
        neighbor_threshold (float): The distance threshold for neighbors.
        remove_intersections (bool): Whether to remove intersecting neighbors.
        log_fn (Callable | None): A function to log messages.

    Returns:
        gdf (gpd.GeoDataFrame): The GeoDataFrame with neighbors.
        injected_neighbor_column_map (InjectedNeighborColumnMap): The injected columns
    """
    injected_neighbor_column_map = InjectedNeighborColumnMap(
        Neighbor_Ixs_col="GLOBI_NEIGHBOR_IXS",
        Neighbor_Polys_col="GLOBI_NEIGHBOR_POLYS",
        Neighbor_Heights_col="GLOBI_NEIGHBOR_HEIGHTS",
        Neighbor_Floors_col="GLOBI_NEIGHBOR_FLOORS",
    )
    log = log_fn or logger.info
    # 2. Compute centroids of the rectangles
    rotated_rectangles = gdf[injected_geometry_col_map.Rotated_Rectangle_col]
    rectangle_centers = rotated_rectangles.apply(lambda x: x.centroid)

    # 3. Extract centroid coordinates
    coords = np.array([
        (geom.x, geom.y) for geom in cast(list[Point], rectangle_centers)
    ])

    # 4. Build a KDTree for efficient spatial queries
    log("building tree")
    tree = cKDTree(coords)
    log("done building tree")

    # 5. Query for neighbors within 50 meters (exclude the point itself)
    indices = tree.query_ball_point(coords, r=neighbor_threshold)
    log("done querying")

    # Remove self from neighbors
    neighbors = [list(set(ind) - {i}) for i, ind in enumerate(indices)]
    log("done removing self")

    # 6. Add the neighbors list to the GeoDataFrame
    gdf[injected_neighbor_column_map.Neighbor_Ixs_col] = neighbors

    def check_row_for_ixs(row: gpd.GeoSeries):
        neighbor_ixs = row[injected_neighbor_column_map.Neighbor_Ixs_col]
        neighbor_rects = gdf.iloc[neighbor_ixs][
            injected_geometry_col_map.Rotated_Rectangle_col
        ]
        current_rect = row[injected_geometry_col_map.Rotated_Rectangle_col]
        intersection = neighbor_rects.apply(lambda x: x.intersects(current_rect))
        non_intersecting_neighbors = pd.Series(
            neighbor_ixs, index=neighbor_rects.index
        )[~intersection].tolist()
        return non_intersecting_neighbors

    if remove_intersections:
        log("removing intersections")
        gdf[injected_neighbor_column_map.Neighbor_Ixs_col] = gdf.apply(
            check_row_for_ixs, axis=1
        )
        log("done removing intersections")
    return gdf, injected_neighbor_column_map


def lon_lat_from_poly(poly: Polygon):
    """Extracts the longitude and latitude coordinates from a given polygon.

    Parameters:
        poly (Polygon): The polygon from which to extract the coordinates.

    Returns:
        lons (list[float]): The longitude coordinates.
        lats (list[float]): The latitude coordinates.
    """
    x, y = poly.exterior.coords.xy
    return x.tolist(), y.tolist()


def extract_neighbors_geo(
    gdf: gpd.GeoDataFrame,
    neighbor_ixs: list[int] | None,
    geometry_col: str = "rotated_rectangle",
) -> list[Polygon | str]:
    """Convert neighbor indices to polygons (or wkt) for a single building.

    Args:
        gdf (gpd.GeoDataFrame): The GeoDataFrame.
        neighbor_ixs (list[int]): The neighbor indices.
        geometry_col (str, optional): The geometry column. Defaults to `rotated_rectangle`.

    Returns:
        list[Polygon | str]: The neighbor polygons.
    """
    if neighbor_ixs is None or len(neighbor_ixs) == 0 or neighbor_ixs is np.nan:
        return []
    polys = gdf.iloc[neighbor_ixs][geometry_col]
    return polys.tolist()


def extract_neighbor_heights(
    gdf: gpd.GeoDataFrame,
    neighbor_ixs: list[int],
    height_col: str = "height",
    fill_na_val: float = 8,
) -> list[float]:
    """Convert neighbor indices to heights for a single building.

    Args:
        gdf (gpd.GeoDataFrame): The GeoDataFrame.
        neighbor_ixs (list[int]): The neighbor indices.
        height_col (str, optional): The height column. Defaults to `height`.
        fill_na_val (float, optional): The value to fill NaNs with. Defaults to 8.

    Returns:
        heights (list[float]): The neighbor heights.
    """
    if neighbor_ixs is None or len(neighbor_ixs) == 0 or neighbor_ixs is np.nan:
        return []
    heights = gdf.iloc[neighbor_ixs][height_col].fillna(fill_na_val)
    return heights.tolist()


def convert_neighbors(
    gdf: gpd.GeoDataFrame,
    neighbor_col: str,
    geometry_col: str,
    height_col: str,
    neighbor_geo_out_col: str,
    neighbor_heights_out_col: str,
    neighbor_floors_out_col: str,
    neighbor_f2f_height: float,
    fill_na_val: float = 2,
) -> gpd.GeoDataFrame:
    """Convert neighbor indices to polygons and heights for all buildings in a GeoDataFrame.

    Args:
        gdf (gpd.GeoDataFrame): The GeoDataFrame.
        neighbor_col (str, optional): The neighbor indices column. Defaults to `neighbor_ixs`.
        geometry_col (str, optional): The geometry column. Defaults to `rotated_rectangle`.
        height_col (str, optional): The height column. Defaults to `height`.
        neighbor_geo_out_col (str, optional): The output column for neighbor polygons. Defaults to `neighbor_polys`.
        neighbor_heights_out_col (str, optional): The output column for neighbor heights. Defaults to `neighbor_heights`.
        neighbor_floors_out_col (str, optional): The output column for neighbor floors. Defaults to `neighbor_floors`.
        neighbor_f2f_height (float, optional): The height of a floor to floor assumption.
        fill_na_val (float, optional): The value to fill NaNs with. Defaults to 2.

    Returns:
        gdf (gpd.GeoDataFrame): The GeoDataFrame
    """
    neighbors = gdf[neighbor_col].apply(
        lambda x: extract_neighbors_geo(
            gdf=gdf,
            neighbor_ixs=x,
            geometry_col=geometry_col,
        )
    )
    neighbor_heights = gdf[neighbor_col].apply(
        lambda x: extract_neighbor_heights(
            gdf=gdf,
            neighbor_ixs=x,
            height_col=height_col,
            fill_na_val=fill_na_val,
        )
    )
    gdf[neighbor_geo_out_col] = neighbors
    gdf[neighbor_heights_out_col] = neighbor_heights
    gdf[neighbor_floors_out_col] = gdf[neighbor_heights_out_col].apply(
        lambda heights: (
            [
                int(h / neighbor_f2f_height)
                if h is not None and not np.isnan(h)
                else None
                for h in heights
            ]
            if isinstance(heights, list)
            else []
        )
    )

    return gdf



"""Type utilities for the GloBI project."""

from globi.type_utils.literals import (
    BasementAtticOccupationConditioningStatus,
    ConditionedOptions,
    OccupiedOptions,
    UnconditionedOptions,
    UnoccupiedOptions,
)

__all__ = [
    "BasementAtticOccupationConditioningStatus",
    "ConditionedOptions",
    "OccupiedOptions",
    "UnconditionedOptions",
    "UnoccupiedOptions",
]



"""Experiment configuration for building builder simulations."""

import logging
import tempfile
from pathlib import Path
from typing import cast

import geopandas as gpd
import numpy as np
import pandas as pd
import yaml
from archetypal.idfclass import IDF
from epinterface.geometry import (
    ShoeboxGeometry,
    compute_shading_mask,
    match_idf_to_building_and_neighbors,
)
from epinterface.sbem.builder import AtticAssumptions, BasementAssumptions, Model
from epinterface.sbem.fields.spec import SemanticModelFields
from numpy.typing import ArrayLike
from pydantic import HttpUrl
from scythe.registry import ExperimentRegistry
from scythe.utils.filesys import FileReference
from shapely import Point, Polygon, from_wkt

from globi.gis.errors import SemanticFieldsFileHasNoBuildingIDColumnError
from globi.gis.geometry import (
    convert_neighbors,
    inject_neighbor_ixs,
    inject_rotated_rectangles,
)
from globi.gis.utils import (
    add_lat_and_lon_cols,
    check_building_ids,
    check_for_column_existence,
    drop_by_area,
    drop_by_edge_length,
    drop_non_polygons,
    handle_attic,
    handle_basement,
    handle_basement_exposed_fraction,
    handle_epwzip,
    handle_height_and_floors,
    handle_wwr,
    inject_semantic_fields,
    rename_shp_cols,
    reproject_gdf,
    validate_has_rows,
    validate_semantic_field_compatibility,
)
from globi.models import (
    DeterministicGISPreprocessorConfig,
    FileConfig,
    GISPreprocessorColumnMap,
    GloBIBuildingSpec,
    GloBIOutputSpec,
)

logger = logging.getLogger(__name__)


INDEX_COLS_TO_KEEP: list[str] = [
    "feature.geometry.long_edge",
    "feature.geometry.short_edge",
    "feature.geometry.orientation",
    "feature.geometry.num_floors",
    "feature.geometry.energy_model_conditioned_area",
    "feature.geometry.energy_model_occupied_area",
    "feature.semantic.Typology",
    "feature.semantic.Age_bracket",
    "feature.semantic.Region",
    "feature.weather.file",
    "feature.geometry.wwr",
    "feature.geometry.f2f_height",
    "feature.geometry.attic_height",
]


def simulate_globi_building_pipeline(
    input_spec: GloBIBuildingSpec,
    tempdir: Path,
) -> GloBIOutputSpec:
    """Simulate a GlobiSpec building and return energy and peak results.

    Args:
        input_spec: The input specification containing building parameters and file URIs
        tempdir: Temporary directory for intermediate files
    Returns:
        Output specification containing a DataFrame with MultiIndex:
        - Top level: Measurement type (Energy, Peak)
        - Feature levels from input specification
    """
    spec = input_spec
    log = logger.info
    zone_def = spec.construct_zone_def()
    model = Model(
        Weather=spec.epwzip_path,
        Zone=zone_def,
        Basement=BasementAssumptions(
            Conditioned=spec.basement_is_conditioned,
            UseFraction=spec.basement_use_fraction
            if spec.basement_is_occupied
            else None,
        ),
        Attic=AtticAssumptions(
            Conditioned=spec.attic_is_conditioned,
            UseFraction=spec.attic_use_fraction if spec.attic_is_occupied else None,
        ),
        geometry=ShoeboxGeometry(
            x=0,
            y=0,
            w=spec.long_edge,
            d=spec.short_edge,
            h=spec.f2f_height,
            wwr=spec.wwr,
            num_stories=spec.num_floors,
            basement=spec.has_basement,
            zoning=spec.use_core_perim_zoning,
            roof_height=spec.attic_height,
            exposed_basement_frac=spec.exposed_basement_frac,
        ),
    )
    # TODO: move this into epinterface.
    azimuthal_angle = 2 * np.pi / 48
    shading_mask = compute_shading_mask(
        spec.rotated_rectangle,
        neighbors=spec.neighbor_polys,
        neighbor_heights=spec.neighbor_heights,
        azimuthal_angle=azimuthal_angle,
    )
    az, p0, p1, h, _w = shading_fence_closed_ring(
        elevations=shading_mask,
        d=100,
    )
    angles = 2 * np.pi * np.arange(len(az)) / len(az)
    p2 = p1 + 2 * np.stack([np.cos(angles), np.sin(angles)], axis=-1)
    p3 = p0 + 2 * np.stack([np.cos(angles), np.sin(angles)], axis=-1)
    rotated_rect_centroid = from_wkt(spec.rotated_rectangle).centroid
    mask_polys = [
        Polygon([
            Point(p0[i] + np.array([rotated_rect_centroid.x, rotated_rect_centroid.y])),
            Point(p1[i] + np.array([rotated_rect_centroid.x, rotated_rect_centroid.y])),
            Point(p2[i] + np.array([rotated_rect_centroid.x, rotated_rect_centroid.y])),
            Point(p3[i] + np.array([rotated_rect_centroid.x, rotated_rect_centroid.y])),
            Point(p0[i] + np.array([rotated_rect_centroid.x, rotated_rect_centroid.y])),
        ])
        for i in range(len(az))
    ]

    def post_geometry_callback(idf: IDF) -> IDF:
        log("Matching IDF to building and neighbors...")
        original_total_building_area = idf.total_building_area
        idf = match_idf_to_building_and_neighbors(
            idf,
            building=spec.rotated_rectangle,
            # neighbor_polys=spec.neighbor_polys,  # pyright: ignore [reportArgumentType]
            # neighbor_floors=spec.neighbor_floors,
            neighbor_polys=mask_polys,  # pyright: ignore [reportArgumentType]
            neighbor_floors=[h[i] // spec.f2f_height for i in range(len(h))],
            neighbor_f2f_height=spec.f2f_height,
            target_short_length=spec.short_edge,
            target_long_length=spec.long_edge,
            rotation_angle=spec.long_edge_angle,
        )
        new_total_building_area = idf.total_building_area
        if not np.isclose(original_total_building_area, new_total_building_area):
            msg = (
                f"Total building area mismatch after matching to building and neighbors: "
                f"{original_total_building_area} != {new_total_building_area}"
            )
            raise ValueError(msg)
        log("IDF matched to building and neighbors.")
        # TODO: possibly consider adding hourly data meter requests to idf here.
        return idf

    log("Building and running model...")
    _idf, results, _err_text, sql, _temp_dir = model.run(
        post_geometry_callback=post_geometry_callback, eplus_parent_dir=tempdir
    )
    # Validate conditioned area
    if not np.allclose(
        model.total_conditioned_area, spec.energy_model_conditioned_area
    ):
        msg = (
            f"Total conditioned area mismatch: "
            f"{model.total_conditioned_area} != {spec.energy_model_conditioned_area}"
        )
        raise ValueError(msg)

    # Results Post-processing
    # TODO: consider if we actually want all t he columns we are including.
    feature_index = spec.make_multiindex(
        n_rows=1, additional_index_data=spec.feature_dict
    )
    results = results.to_frame().T.set_index(feature_index)

    dfs: dict[str, pd.DataFrame] = {
        "Results": results,
    }

    hourly_data_outpath: FileReference | None = None

    if spec.parent_experiment_spec and spec.parent_experiment_spec.hourly_data_config:
        hourly_df = sql.timeseries_by_name(
            spec.parent_experiment_spec.hourly_data_config.data,
            reporting_frequency="Hourly",
        )
        hourly_df.index.names = ["Timestep"]
        hourly_df.columns.names = ["Trash", "Zone", "Meter"]
        hourly_df: pd.DataFrame = cast(
            pd.DataFrame,
            hourly_df.droplevel("Trash", axis=1)
            .stack(level="Zone", future_stack=True)
            .unstack(level="Timestep"),
        )
        hourly_multiindex = spec.make_multiindex(
            n_rows=len(hourly_df), include_sort_subindex=False
        )
        old_ix = hourly_df.index
        hourly_df.index = hourly_multiindex
        hourly_df = hourly_df.set_index(old_ix, append=True)

        if spec.parent_experiment_spec.hourly_data_config.does_dataframe_output:
            dfs["HourlyData"] = hourly_df
        if spec.parent_experiment_spec.hourly_data_config.does_file_output:
            hourly_data_outpath = tempdir / "outputs_hourly_data.pq"
            hourly_df.to_parquet(hourly_data_outpath)

    return GloBIOutputSpec(
        dataframes=dfs,
        hourly_data=hourly_data_outpath,
    )


@ExperimentRegistry.Register(retries=2, schedule_timeout="10h", execution_timeout="30m")
def simulate_globi_building(
    input_spec: GloBIBuildingSpec, tempdir: Path
) -> GloBIOutputSpec:
    """Simulate a GlobiSpec building and return monthly energy and peak results.

    NB: this is separated from the pipeline above so the pipeline can still be used as a
    local invocation without *too* much difficulty.
    """
    return simulate_globi_building_pipeline(input_spec, tempdir)


if __name__ == "__main__":
    from globi.models import FileConfig, GloBIExperimentSpec, HourlyDataConfig

    region = "Brazil"
    data_dir = Path(__file__).parent.parent.parent / "data"
    db_path = data_dir / region / "components-lib.db"
    semantic_fields_path = data_dir / region / "semantic-fields.yaml"
    component_map_path = data_dir / region / "component-map.yaml"
    weather_url = "https://climate.onebuilding.org/WMO_Region_3_South_America/BRA_Brazil/SP_Sao_Paulo/BRA_SP_Guaratingueta.AP.837080_TMYx.2009-2023.zip"
    epwzip_file = HttpUrl(weather_url)
    globi_spec = GloBIBuildingSpec(
        experiment_id="brazil-test-experiment",
        db_file=db_path,
        semantic_fields_file=semantic_fields_path,
        component_map_file=component_map_path,
        epwzip_file=epwzip_file,
        long_edge=20,
        short_edge=15,
        long_edge_angle=0,
        aspect_ratio=20 / 15,
        rotated_rectangle_area_ratio=1.0,
        num_floors=3,
        height=3 * 3.0,
        f2f_height=3.0,
        wwr=0.2,
        rotated_rectangle="POLYGON ((0 0, 0 15, 20 15, 20 0, 0 0))",
        neighbor_polys=[],
        neighbor_floors=[],
        neighbor_heights=[],
        semantic_field_context={
            "region": "SP",
            "typology": "Residential",
            "income": "Medium",
            "scenario": "noAC",
        },
        basement="none",
        attic="none",
        exposed_basement_frac=0.25,
        sort_index=0,
        parent_experiment_spec=GloBIExperimentSpec(
            name="placeholder",
            hourly_data_config=HourlyDataConfig(
                data=("Zone Mean Air Temperature",),
                output_mode="dataframes-only",
            ),
            file_config=FileConfig(
                gis_file=Path("placeholder"),
                db_file=db_path,
                semantic_fields_file=semantic_fields_path,
                epwzip_file=str(epwzip_file),
                component_map_file=component_map_path,
            ),
            scenario="test",
        ),
    )

    print("Test simulation...")
    with tempfile.TemporaryDirectory() as tempdir:
        r = simulate_globi_building_pipeline(globi_spec, tempdir=Path(tempdir))


def preprocess_gis_file(
    config: DeterministicGISPreprocessorConfig,
    file_config: "FileConfig",
    scenario: str | None = None,
    output_path: Path | None = None,
    load_from_output_if_present: bool = True,
) -> tuple[gpd.GeoDataFrame, GISPreprocessorColumnMap]:
    """Preprocess a GIS file.

    Args:
        config (DeterministicGISPreprocessorConfig): The configuration for the GIS preprocessor.
        file_config (FileConfig): The configuration for the files.
        scenario (str | None): The scenario identifier to add to the semantic field context.
        output_path (Path | None): Optional folder to save preprocessed files. If None, file is not saved.
        load_from_output_if_present (bool): If True, loads the preprocessed file from the output path if it exists.

    Returns:
        gdf (gpd.GeoDataFrame): The preprocessed GeoDataFrame.
    """
    if output_path is not None and output_path.is_file():
        msg = f"Expected a folder, got {output_path}"
        raise ValueError(msg)
    gis_fp = file_config.gis_file
    if load_from_output_if_present and output_path is not None:
        gdf_output_path = output_path / "globi_gdf.pq"
        column_output_map_output_path = output_path / "globi_column_output_map.yaml"
        gdf = cast(gpd.GeoDataFrame, gpd.read_parquet(gdf_output_path))
        column_output_map = GISPreprocessorColumnMap.from_manifest(
            column_output_map_output_path
        )
        return gdf, column_output_map

    # load the semantic fields
    # We will need to access this as it stores some
    # rich information about which columns in the provided GIS data will
    # contain standard provided values, like wwr, height etc etc.
    # it also stores the fields that will be used for semantic mapping,
    # so we can run a consistency check with the component map.
    with open(file_config.semantic_fields_file) as f:
        semantic_fields = SemanticModelFields.model_validate(yaml.safe_load(f))
    if semantic_fields.Building_ID_col is None:
        raise SemanticFieldsFileHasNoBuildingIDColumnError()

    gdf = cast(gpd.GeoDataFrame, gpd.read_file(gis_fp))

    validate_has_rows(gdf)

    # Check that the current CRS is WGS84 or the cart one, convert to WGS84 early and use throughout
    gdf = reproject_gdf(gdf, config.cart_crs)

    required_col_names = semantic_fields.field_names

    # We need to deal with the fact that shapefiles will trucnate the column
    # name to 10 characters, but users might not realize this when they
    # export from e.g. ArcGIS.
    gdf = rename_shp_cols(gdf, required_col_names, log_fn=logger.info)
    if scenario is not None:
        gdf["scenario"] = scenario

    # We want to run a consistency check to make sure that the requested semantic fields
    # are actually in the GDF after we have dealt with appropriate renaming.
    # We also should run a consistency check to make sure that every cell value that is listed as a
    # semantic field is actually one of the expected values.
    check_for_column_existence(gdf, required_col_names, log_fn=logger.info)
    validate_semantic_field_compatibility(
        gdf,
        semantic_fields,
        missing_ok=False,
        log_fn=logger.info,
    )

    # If the building ID column is not provided or partial, we will inject uuids
    gdf, semantic_fields.Building_ID_col = check_building_ids(
        gdf, semantic_fields.Building_ID_col, log_fn=logger.info
    )

    # We add the latitude and longitude columns to the GeoDataFrame.
    gdf = add_lat_and_lon_cols(gdf)

    # We store the initial building count since we will now start dropping rows.
    initial_count = len(gdf)

    # We deal with imputing heights/number of floors depending on what is present.
    (
        gdf,
        semantic_fields.Height_col,
        semantic_fields.Num_Floors_col,
        f2f_height_col,
        n_dropped_by_floors_heights,
    ) = handle_height_and_floors(
        gdf,
        height_col=semantic_fields.Height_col,
        nfloors_col=semantic_fields.Num_Floors_col,
        assumed_f2f_height=config.f2f_height,
        default_n_floors=config.default_num_floors,
        min_floors=config.min_num_floors,
        max_floors=config.max_num_floors,
        min_height=config.min_building_height,
        max_height=config.max_building_height,
        log_fn=logger.info,
    )
    validate_has_rows(gdf)

    # Next, we deal with imputing various rich columns.
    (
        gdf,
        semantic_fields.WWR_col,
        n_dropped_by_wwr,
    ) = handle_wwr(
        gdf,
        wwr_col=semantic_fields.WWR_col,
        assumed_wwr=config.default_wwr,
        log_fn=logger.info,
    )
    validate_has_rows(gdf)
    (
        gdf,
        semantic_fields.Basement_col,
        n_dropped_by_basement,
    ) = handle_basement(
        gdf,
        basement_col=semantic_fields.Basement_col,
        assumed_basement=config.default_basement,
        log_fn=logger.info,
    )
    validate_has_rows(gdf)
    (
        gdf,
        semantic_fields.Exposed_Basement_Frac_col,
        n_dropped_by_basement_exposed_fraction,
    ) = handle_basement_exposed_fraction(
        gdf,
        basement_exposed_fraction_col=semantic_fields.Exposed_Basement_Frac_col,
        assumed_basement_exposed_fraction=config.default_exposed_basement_frac,
        log_fn=logger.info,
    )
    validate_has_rows(gdf)
    (
        gdf,
        semantic_fields.Attic_col,
        n_dropped_by_attic,
    ) = handle_attic(
        gdf,
        attic_col=semantic_fields.Attic_col,
        assumed_attic=config.default_attic,
        log_fn=logger.info,
    )
    validate_has_rows(gdf)
    gdf, n_dropped_by_points = drop_non_polygons(gdf, log_fn=logger.info)
    validate_has_rows(gdf)
    logger.info("injecting rotated rectangles")
    gdf, injected_geometry_column_map = inject_rotated_rectangles(
        gdf, cart_crs=config.cart_crs
    )
    gdf, n_dropped_by_area = drop_by_area(
        gdf,
        area_col=injected_geometry_column_map.Footprint_Area_col,
        min_area=config.min_building_area,
        log_fn=logger.info,
    )
    validate_has_rows(gdf)
    gdf, n_dropped_by_edge = drop_by_edge_length(
        gdf,
        short_edge_col=injected_geometry_column_map.Short_Edge_col,
        long_edge_col=injected_geometry_column_map.Long_Edge_col,
        min_edge_length_m=config.min_edge_length,
        max_edge_length_m=config.max_edge_length,
        log_fn=logger.info,
    )

    n_dropped = (
        n_dropped_by_area
        + n_dropped_by_edge
        + n_dropped_by_floors_heights
        + n_dropped_by_wwr
        + n_dropped_by_basement_exposed_fraction
        + n_dropped_by_basement
        + n_dropped_by_attic
        + n_dropped_by_points
    )
    logger.info(f"Dropped {n_dropped / initial_count:.1%} of all buildings.")

    validate_has_rows(gdf)

    logger.info(
        f"Retained {len(gdf)} buildings after filtering (removed {initial_count - len(gdf)} total)"
    )

    logger.info("computing neighbor indices")
    gdf, injected_neighbor_column_map = inject_neighbor_ixs(
        cast(gpd.GeoDataFrame, gdf),
        injected_geometry_col_map=injected_geometry_column_map,
        neighbor_threshold=config.neighbor_threshold,
        log_fn=logger.info,
    )

    logger.info("extracting and converting neighbors")
    gdf = convert_neighbors(
        gdf,
        neighbor_col=injected_neighbor_column_map.Neighbor_Ixs_col,
        geometry_col=injected_geometry_column_map.Rotated_Rectangle_col,
        height_col=semantic_fields.Height_col,
        neighbor_geo_out_col=injected_neighbor_column_map.Neighbor_Polys_col,
        neighbor_heights_out_col=injected_neighbor_column_map.Neighbor_Heights_col,
        neighbor_floors_out_col=injected_neighbor_column_map.Neighbor_Floors_col,
        fill_na_val=config.default_num_floors * config.f2f_height,
        neighbor_f2f_height=config.f2f_height,
    )

    # Construct a dictionary of the semantic field values for each building.
    gdf, semantic_fields_context_col = inject_semantic_fields(gdf, semantic_fields)

    # EPW FILE HANDLING
    gdf, semantic_fields.Weather_File_col = handle_epwzip(
        gdf,
        weather_file_col=semantic_fields.Weather_File_col,
        assumed_epwzip=file_config.epwzip_file,
        epw_query=config.epw_query,
        cart_crs=config.cart_crs,
        log_fn=logger.info,
    )

    db_file_col = "GLOBI_DB_FILE"
    semantic_fields_file_col = "GLOBI_SEMANTIC_FIELDS_FILE"
    component_map_file_col = "GLOBI_COMPONENT_MAP_FILE"
    gdf[db_file_col] = file_config.db_file
    gdf[semantic_fields_file_col] = file_config.semantic_fields_file
    gdf[component_map_file_col] = file_config.component_map_file

    column_output_map = GISPreprocessorColumnMap(
        DB_File_col=db_file_col,
        Semantic_Fields_File_col=semantic_fields_file_col,
        Component_Map_File_col=component_map_file_col,
        EPWZip_File_col=semantic_fields.Weather_File_col,
        Semantic_Field_Context_col=semantic_fields_context_col,
        Neighbor_Polys_col=injected_neighbor_column_map.Neighbor_Polys_col,
        Neighbor_Heights_col=injected_neighbor_column_map.Neighbor_Heights_col,
        Neighbor_Floors_col=injected_neighbor_column_map.Neighbor_Floors_col,
        Rotated_Rectangle_col=injected_geometry_column_map.Rotated_Rectangle_col,
        Long_Edge_Angle_col=injected_geometry_column_map.Long_Edge_Angle_col,
        Long_Edge_col=injected_geometry_column_map.Long_Edge_col,
        Short_Edge_col=injected_geometry_column_map.Short_Edge_col,
        Aspect_Ratio_col=injected_geometry_column_map.Aspect_Ratio_col,
        Rotated_Rectangle_Area_Ratio_col=injected_geometry_column_map.Rotated_Rectangle_Area_Ratio_col,
        WWR_col=semantic_fields.WWR_col,
        Height_col=semantic_fields.Height_col,
        Num_Floors_col=semantic_fields.Num_Floors_col,
        F2F_Height_col=f2f_height_col,
        Basement_col=semantic_fields.Basement_col,
        Attic_col=semantic_fields.Attic_col,
        Exposed_Basement_Frac_col=semantic_fields.Exposed_Basement_Frac_col,
    )

    # TODO: make sure we can save the file still

    if output_path is not None:
        # TODO: make sure path is correct, geojson, etc.
        logger.info(f"saving preprocessed gis file to: {output_path}")
        gdf_output_path = output_path / "globi_gdf.pq"
        column_output_map_output_path = output_path / "globi_column_output_map.yaml"
        gdf.to_parquet(gdf_output_path)
        with open(column_output_map_output_path, "w") as f:
            yaml.dump(
                column_output_map.model_dump(mode="json"), f, sort_keys=False, indent=2
            )
        logger.info(f"saved {len(gdf)} features to {output_path}")

    return gdf, column_output_map
    # print("\nOutput dataframes:", list(output.dataframes.keys()))
    # results_df = output.dataframes["Results"]
    # print("Results shape:", results_df.shape)
    # print("Results index levels:", results_df.index.names)
    # print("\nEnergy data:")
    # print(results_df.loc["Energy"])
    # print("\nPeak data:")
    # print(results_df.loc["Peak"])


# TODO: move to epinterface
def shading_fence_closed_ring(
    elevations: ArrayLike,
    d: float,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, float]:
    """Construct N vertical 'shading fence' rectangles whose bases are tangent segments forming a closed regular polygon around a circle of radius d.

    Inputs:
      elevations: (N,) elevation angles theta_k [radians]
      d: radius to tangency points (base midpoints)

    Outputs:
      azimuths: (N,) inferred azimuths alpha_k = 2πk/N
      p0:       (N, 2) base endpoint A (x,y)
      p1:       (N, 2) base endpoint B (x,y)
      h:        (N,) heights h_k = d * tan(theta_k)
      w:        scalar side length / segment width = 2 d tan(π/N)

    Notes:
      - With this construction, the segments intersect/meet: p1[k] == p0[k+1] (cyclic),
        up to floating point tolerance.
      - N must be >= 3 for a closed polygon.
    """
    theta = np.asarray(elevations, dtype=np.float64)
    if theta.ndim != 1:
        msg = f"elevations must be 1D, got shape {theta.shape}"
        raise ValueError(msg)

    N = theta.shape[0]
    if N < 3:
        msg = "Need at least 3 elevations (N >= 3) to form a closed ring."
        raise ValueError(msg)

    # Inferred equally spaced azimuths
    k = np.arange(N, dtype=np.float64)
    azimuths = 2.0 * np.pi * k / N

    ca, sa = np.cos(azimuths), np.sin(azimuths)

    # Tangency points (base midpoints) on circle radius d
    px = d * ca
    py = d * sa

    # Unit tangent direction (perpendicular to radius)
    tx = -sa
    ty = ca

    # Choose width so adjacent tangent segments meet (circumscribed regular N-gon)
    half_w = d * np.tan(np.pi / N)
    w = 2.0 * half_w

    # Endpoints in xy: p ± half_w * t
    p0 = np.stack([px - half_w * tx, py - half_w * ty], axis=-1)
    p1 = np.stack([px + half_w * tx, py + half_w * ty], axis=-1)

    # Heights from elevation angles
    h = d * np.tan(theta)

    return azimuths, p0, p1, h, w


if __name__ == "__main__":
    from globi.models import GloBIExperimentSpec

    spec = GloBIExperimentSpec.from_(Path("data/models/Cambridge_UK/manifest.yml"))
    print(yaml.dump(spec.model_dump(mode="json"), indent=2, sort_keys=False))
    # gdf = gpd.read_file(spec.file_config.gis_file)
    logging.basicConfig(level=logging.DEBUG)
    if spec.gis_preprocessor_config is not None:
        new_gdf_path = (
            spec.file_config.gis_file.parent / "buildings-seed-updated.geojson"
        )
        old_gdf: gpd.GeoDataFrame = gpd.read_file(spec.file_config.gis_file)
        old_gdf["Region"] = "CB"
        old_gdf["Age_bracket"] = (
            old_gdf["Age_bracket"].str.replace(" ", "_").str.replace("1999", "1960")
        )
        old_gdf["OCCUPANCY_DENSITY"] = old_gdf["OCCUPANCY_DENSITY"].str.replace(
            "MediumDensity", "MedDensity"
        )
        old_gdf.to_file(new_gdf_path, driver="GeoJSON")
        spec.file_config.gis_file = new_gdf_path
        gdf, colmap = preprocess_gis_file(
            config=spec.gis_preprocessor_config,
            file_config=spec.file_config,
        )
        print(gdf.columns)
        print(colmap.model_dump_json(indent=2))
        print(gdf.head())
        import pandas as pd

        print(cast(pd.Series, gdf[colmap.EPWZip_File_col]).unique())



"""A model for an SBEM composed shoebox simulation."""

import logging
import tempfile
from functools import cached_property
from pathlib import Path
from typing import Annotated, Literal, Self, cast

import numpy as np
import yaml
from epinterface.geometry import compute_shading_mask
from epinterface.sbem.components.composer import (
    construct_composer_model,
    construct_graph,
)
from epinterface.sbem.components.zones import ZoneComponent
from epinterface.sbem.prisma.client import PrismaSettings
from pydantic import BaseModel, BeforeValidator, Field
from scythe.base import ExperimentInputSpec, ExperimentOutputSpec
from scythe.utils.filesys import FileReference, fetch_uri

from globi.type_utils import (
    BasementAtticOccupationConditioningStatus,
    ConditionedOptions,
    OccupiedOptions,
)

logger = logging.getLogger(__name__)


AvailableHourlyData = Literal[
    "Zone Mean Air Temperature",
    "Zone Air Relative Humidity",
]


class BaseConfig(BaseModel):
    """A base configuration for a Globi experiment."""

    @classmethod
    def from_manifest(cls, manifest_path: Path) -> Self:
        """Load the base configuration from a manifest file."""
        with open(manifest_path) as f:
            manifest = yaml.safe_load(f)
        return cls.model_validate(manifest)

    @classmethod
    def from_manifest_fileref(cls, manifest_fileref: FileReference) -> Self:
        """Load the base configuration from a manifest file reference."""
        if isinstance(manifest_fileref, str) and (
            not manifest_fileref.startswith("http://")
            and not manifest_fileref.startswith("https://")
            and not manifest_fileref.startswith("s3://")
        ):
            manifest_fileref = Path(manifest_fileref)
        if isinstance(manifest_fileref, Path):
            local_path = manifest_fileref
        else:
            with tempfile.TemporaryDirectory() as temp_dir:
                temp_path = Path(temp_dir) / "manifest.yaml"
                local_path = fetch_uri(manifest_fileref, temp_path)
        return cls.from_manifest(local_path)

    @classmethod
    def from_(cls, v: Self | FileReference) -> Self:
        """Load the base configuration from a manifest file reference or a local path."""
        if isinstance(v, BaseConfig | dict):
            return v
        return cls.from_manifest_fileref(v)


class HourlyDataConfig(BaseConfig):
    """Configuration for hourly data."""

    data: tuple[AvailableHourlyData, ...] = Field(
        default=(),
        description="The hourly data to report.",
    )

    output_mode: Literal[
        "dataframes-and-filerefs", "fileref-only", "dataframes-only"
    ] = Field(
        default="dataframes-and-filerefs",
        description="The mode to output the hourly data.",
    )

    @property
    def does_file_output(self) -> bool:
        """Whether the hourly data output is a file."""
        return self.output_mode in ["dataframes-and-filerefs", "fileref-only"]

    @property
    def does_dataframe_output(self) -> bool:
        """Whether the hourly data output is a dataframe."""
        return self.output_mode in ["dataframes-and-filerefs", "dataframes-only"]


class DeterministicGISPreprocessorConfig(BaseConfig):
    """Configuration for the GIS preprocessor."""

    # TODO: design decision - Separated out this config since this would be for deterministic elements primarily

    cart_crs: str = Field(
        default="EPSG:3857",
        description="The cartesian CRS to project to.",
    )
    min_building_area: float = Field(
        default=10.0,
        ge=1,
        le=1000,
        description="The minimum area of a building to be included [m^2].",
    )
    min_edge_length: float = Field(
        default=3.0,
        ge=1,
        le=2000,
        description="The minimum edge length of a building to be included [m].",
    )
    max_edge_length: float = Field(
        default=1000.0,
        ge=1,
        le=2000,
        description="The maximum edge length of a building to be included [m].",
    )
    neighbor_threshold: float = Field(
        default=100.0,
        ge=0,
        description="The distance threshold for neighbors [m].",
    )
    f2f_height: float = Field(
        default=3.0,
        ge=2,
        le=5,
        description="The floor-to-floor height [m].",
    )
    min_building_height: float = Field(
        default=3,
        ge=1,
        le=500,
        description="The minimum building height [m].",
    )
    max_building_height: float = Field(
        default=300,
        ge=1,
        le=500,
        description="The maximum building height [m].",
    )
    min_num_floors: int = Field(
        default=1,
        ge=1,
        le=150,
        description="The minimum number of floors.",
    )
    max_num_floors: int = Field(
        default=125,
        ge=1,
        le=150,
        description="The maximum number of floors.",
    )

    default_wwr: float = Field(
        default=0.2,
        ge=0,
        le=1,
        description="The default window-to-wall ratio.",
    )
    default_num_floors: int = Field(
        default=2,
        ge=1,
        description="The default number of floors.",
    )
    default_basement: BasementAtticOccupationConditioningStatus = Field(
        default_factory=lambda: "none", description="The default basement type."
    )
    default_attic: BasementAtticOccupationConditioningStatus = Field(
        default_factory=lambda: "none", description="The default attic type."
    )
    default_exposed_basement_frac: float = Field(
        default=0.25,
        ge=0,
        le=1,
        description="The default exposed basement fraction.",
    )
    epw_query: str | None = Field(
        default_factory=lambda: "source in ['tmyx']",
        description="The EPW query filter for closest_epw.",
    )


class GISPreprocessorColumnMap(BaseConfig):
    """Output for the GIS preprocessor column names."""

    DB_File_col: str
    Semantic_Fields_File_col: str
    Component_Map_File_col: str
    EPWZip_File_col: str
    Semantic_Field_Context_col: str
    Neighbor_Polys_col: str
    Neighbor_Heights_col: str
    Neighbor_Floors_col: str
    Rotated_Rectangle_col: str
    Long_Edge_Angle_col: str
    Long_Edge_col: str
    Short_Edge_col: str
    Aspect_Ratio_col: str
    Rotated_Rectangle_Area_Ratio_col: str
    WWR_col: str
    Height_col: str
    Num_Floors_col: str
    F2F_Height_col: str
    Basement_col: str
    Attic_col: str
    Exposed_Basement_Frac_col: str


class FileConfig(BaseConfig):
    """Configuration for files."""

    gis_file: Path = Field(..., description="The path to the local GIS file.")
    db_file: Path  # these could be file refs?
    semantic_fields_file: Path  # these could be file refs?
    epwzip_file: (
        Path | str | None
    )  # TODO: our gis to model conversion should handle EPW identification; see gis job submission in epengine
    component_map_file: Path


ReferencedHourlyDataConfig = Annotated[
    HourlyDataConfig, BeforeValidator(HourlyDataConfig.from_)
]
ReferencedFileConfig = Annotated[FileConfig, BeforeValidator(FileConfig.from_)]
ReferencedGISPreprocessorConfig = Annotated[
    DeterministicGISPreprocessorConfig,
    BeforeValidator(DeterministicGISPreprocessorConfig.from_),
]


class GloBIExperimentSpec(BaseConfig):
    """Specification for a Globi experiment."""

    name: str = Field(..., description="The name of the experiment.")
    scenario: str = Field(..., description="The scenario identifier.")
    hourly_data_config: ReferencedHourlyDataConfig | None = Field(
        default=None,
        description="The configuration for the hourly data.",
    )
    file_config: ReferencedFileConfig = Field(
        ..., description="The configuration for the files."
    )
    gis_preprocessor_config: ReferencedGISPreprocessorConfig = Field(
        default_factory=DeterministicGISPreprocessorConfig,
        description="The configuration for the GIS preprocessor.",
    )


class GloBIBuildingSpec(ExperimentInputSpec):
    """A spec for running an EnergyPlus simulation for any region."""

    # TODO: update the nullability
    db_file: FileReference = Field(..., description="The component database file.")
    semantic_fields_file: FileReference = Field(
        ..., description="The semantic fields file."
    )
    component_map_file: FileReference = Field(
        ..., description="The component map file."
    )
    epwzip_file: FileReference = Field(..., description="The EPW weather file.")
    semantic_field_context: dict[str, float | str | int] = Field(
        ...,
        description="The semantic field values which will be used to compile the zone definition.",
    )
    neighbor_polys: list[str] = Field(
        ..., description="The polygons of the neighboring buildings."
    )
    neighbor_heights: list[float | int | None] = Field(
        ..., description="The height of the neighboring buildings  [m]."
    )
    neighbor_floors: list[float | int | None] = Field(
        ..., description="The number of floors of the neighboring buildings."
    )
    rotated_rectangle: str = Field(
        ..., description="The rotated rectangle fitted around the base of the building."
    )
    long_edge_angle: float = Field(
        ..., description="The long edge angle of the building (radians)."
    )
    long_edge: float = Field(
        ..., description="The length of the long edge of the building [m]."
    )
    short_edge: float = Field(
        ..., description="The length of the short edge of the building [m]."
    )
    aspect_ratio: float = Field(
        ..., description="The aspect ratio of the building footprint [unitless]."
    )
    rotated_rectangle_area_ratio: float = Field(
        ...,
        description="The ratio of the rotated rectangle footprint area to the building footprint area.",
    )
    wwr: float = Field(
        ...,
        description="The window-to-wall ratio of the building [unitless].",
        ge=0,
        le=1,
    )
    height: float = Field(..., description="The height of the building [m].", ge=0)
    num_floors: int = Field(
        ..., description="The number of floors in the building.", ge=0
    )
    f2f_height: float = Field(..., description="The floor to floor height [m].", ge=0)
    basement: BasementAtticOccupationConditioningStatus = Field(
        ..., description="The type of basement in the building."
    )
    attic: BasementAtticOccupationConditioningStatus = Field(
        ..., description="The type of attic in the building."
    )
    exposed_basement_frac: float = Field(
        ...,
        description="The fraction of the basement that is exposed to the air.",
        gt=0,
        lt=1,
    )

    parent_experiment_spec: GloBIExperimentSpec | None = Field(
        default=None,
        description="The parent experiment spec.",
    )

    @property
    def feature_dict(self) -> dict[str, str | int | float]:
        """Return a dictionary of features which will be available to ML algos."""
        features: dict[str, str | int | float] = {
            "feature.geometry.long_edge": self.long_edge,
            "feature.geometry.short_edge": self.short_edge,
            "feature.geometry.orientation": self.long_edge_angle,
            "feature.geometry.orientation.cos": np.cos(self.long_edge_angle),
            "feature.geometry.orientation.sin": np.sin(self.long_edge_angle),
            "feature.geometry.aspect_ratio": self.aspect_ratio,
            "feature.geometry.wwr": self.wwr,
            "feature.geometry.num_floors": self.num_floors,
            "feature.geometry.f2f_height": self.f2f_height,
            # "feature.geometry.fp_area": self.fp_area,
            "feature.geometry.zoning": self.use_core_perim_zoning,
            "feature.geometry.energy_model_conditioned_area": self.energy_model_conditioned_area,
            "feature.geometry.energy_model_occupied_area": self.energy_model_occupied_area,
            "feature.geometry.attic_height": self.attic_height or 0,
            "feature.geometry.exposed_basement_frac": self.exposed_basement_frac,
        }

        # TODO: consider passing in
        # neighbors directly to Model.geometry, letting model perform neighbor
        # insertion directly rather than via a callback,
        # and then let shading mask become a computed property of the model.geometry.
        shading_mask = compute_shading_mask(
            self.rotated_rectangle,
            neighbors=self.neighbor_polys,
            neighbor_heights=self.neighbor_heights,
            azimuthal_angle=2 * np.pi / 48,
        )
        shading_mask_values = {
            f"feature.geometry.shading_mask_{i:02d}": val
            for i, val in enumerate(shading_mask.tolist())
        }
        features.update(shading_mask_values)

        # semantic features are kept separately as one building may have
        # multiple simulations with different semantic fields.
        features.update({
            f"feature.semantic.{feature_name}": feature_value
            for feature_name, feature_value in self.semantic_field_context.items()
        })

        features["feature.weather.file"] = self.epwzip_path.stem

        # conditional features are derived from the static and semantic features,
        # and may be subject to things like conditional sampling, estimation etc.
        # e.g. rvalues, uvalues, schedule, etc.
        # additional things like basement/attic config?
        features["feature.extra_spaces.basement.exists"] = (
            "Yes" if self.has_basement else "No"
        )
        features["feature.extra_spaces.basement.occupied"] = (
            "Yes" if self.basement_is_occupied else "No"
        )
        features["feature.extra_spaces.basement.conditioned"] = (
            "Yes" if self.basement_is_conditioned else "No"
        )
        features["feature.extra_spaces.basement.use_fraction"] = (
            self.basement_use_fraction
        )
        features["feature.extra_spaces.attic.exists"] = (
            "Yes" if self.has_attic else "No"
        )
        features["feature.extra_spaces.attic.occupied"] = (
            "Yes" if self.attic_is_occupied else "No"
        )
        features["feature.extra_spaces.attic.conditioned"] = (
            "Yes" if self.attic_is_conditioned else "No"
        )
        features["feature.extra_spaces.attic.use_fraction"] = self.attic_use_fraction

        return features

    # TODO: use the scythe automatic referencing for these paths - FileReference class from scythe.utils.files
    # choose a local file and direclty use the 'Path' for this
    # self scythe - fetch uri
    # input_sepc.weather_file
    # everything gets a tempdir

    #

    @cached_property
    def db_path(self) -> Path:
        """Fetch the db file and return the local path.

        Returns:
            local_path (Path): The local path of the db file
        """
        if isinstance(self.db_file, Path):
            return self.db_file
        return self.fetch_uri(self.db_file)

    @cached_property
    def semantic_fields_path(self) -> Path:
        """Fetch the semantic fields file and return the local path.

        Returns:
            local_path (Path): The local path of the semantic fields file
        """
        if isinstance(self.semantic_fields_file, Path):
            return self.semantic_fields_file
        return self.fetch_uri(self.semantic_fields_file)

    @cached_property
    def epwzip_path(self) -> Path:
        """Fetch the epw file and return the local path.

        Returns:
            local_path (Path): The local path of the epw file
        """
        if isinstance(self.epwzip_file, Path):
            return self.epwzip_file
        return self.fetch_uri(self.epwzip_file)

    @property
    def component_map(self) -> Path:
        """Fetch the component map file and return the local path.

        Returns:
            local_path (Path): The local path of the component map file
        """
        if isinstance(self.component_map_file, Path):
            return self.component_map_file
        return self.fetch_uri(self.component_map_file)

    def construct_zone_def(self) -> ZoneComponent:
        """Construct the zone definition for the simulation.

        Returns:
            zone_def (ZoneComponent): The zone definition for the simulation
        """
        g = construct_graph(ZoneComponent)
        SelectorModel = construct_composer_model(
            g,
            ZoneComponent,
            use_children=False,
        )

        with open(self.component_map) as f:
            component_map_yaml = yaml.safe_load(f)
        selector = SelectorModel.model_validate(component_map_yaml)

        # Log the database path being used for debugging
        import os
        from datetime import datetime

        if self.db_path.exists():
            mtime = os.path.getmtime(self.db_path)
            mtime_str = datetime.fromtimestamp(mtime).strftime("%Y-%m-%d %H:%M:%S")
            logger.info(
                f"Loading database: {self.db_path} "
                f"(modified: {mtime_str}, size: {self.db_path.stat().st_size} bytes)"
            )
        else:
            logger.error(f"Database file not found: {self.db_path}")

        # Force a fresh database connection by creating a new PrismaSettings instance
        # This ensures we always reload the database, avoiding any caching issues.
        # Each call to construct_zone_def() creates a new PrismaSettings instance,
        # which should force SQLite to open a fresh connection and see any file updates.
        settings = PrismaSettings.New(
            database_path=self.db_path, if_exists="ignore", auto_register=False
        )
        db = settings.db

        # Use context manager to ensure connection is properly closed after use.
        # This ensures SQLite releases file locks and any future reads will see
        # updated database content.
        context = self.semantic_field_context
        with db:
            zone = cast(ZoneComponent, selector.get_component(context=context, db=db))
        # Connection is now closed, ensuring any future reads will see updated data
        return zone

    @property
    def use_core_perim_zoning(self) -> Literal["by_storey", "core/perim"]:
        """Whether to use the core perimeter for the simulation."""
        use_core_perim = self.long_edge > 15 and self.short_edge > 15
        return "core/perim" if use_core_perim else "by_storey"

    @property
    def basement_is_occupied(self) -> bool:
        """Whether the basement is occupied."""
        return self.basement in OccupiedOptions

    @property
    def attic_is_occupied(self) -> bool:
        """Whether the attic is occupied."""
        return self.attic in OccupiedOptions

    @property
    def basement_is_conditioned(self) -> bool:
        """Whether the basement is conditioned."""
        return self.basement in ConditionedOptions

    @property
    def attic_is_conditioned(self) -> bool:
        """Whether the attic is conditioned."""
        return self.attic in ConditionedOptions

    @cached_property
    def basement_use_fraction(self) -> float:
        """The use fraction of the basement."""
        if not self.basement_is_occupied:
            return 0
        return np.random.uniform(0.2, 0.6)

    @cached_property
    def attic_use_fraction(self) -> float:
        """The use fraction of the attic."""
        if not self.attic_is_occupied:
            return 0
        # TODO: use sampling as a fallback value when a default is not provided rather
        # than always sampling.
        return np.random.uniform(0.2, 0.6)

    @cached_property
    def has_basement(self) -> bool:
        """Whether the building has a basement."""
        return self.basement != "none"

    @cached_property
    def has_attic(self) -> bool:
        """Whether the building has an attic."""
        return self.attic != "none"

    @cached_property
    def attic_height(self) -> float | None:
        """The height of the attic."""
        if not self.has_attic:
            return None
        min_occupied_or_conditioned_rise_over_run = 6 / 12
        max_occupied_or_conditioned_rise_over_run = 9 / 12
        min_unoccupied_and_unconditioned_rise_over_run = 4 / 12
        max_unoccupied_and_unconditioned_rise_over_run = 6 / 12

        run = self.short_edge / 2
        attic_height = None
        attempts = 20
        while attic_height is None and attempts > 0:
            if self.attic_is_occupied or self.attic_is_conditioned:
                attic_height = run * np.random.uniform(
                    min_occupied_or_conditioned_rise_over_run,
                    max_occupied_or_conditioned_rise_over_run,
                )
            else:
                attic_height = run * np.random.uniform(
                    min_unoccupied_and_unconditioned_rise_over_run,
                    max_unoccupied_and_unconditioned_rise_over_run,
                )
            if attic_height > self.f2f_height * 2.5:
                attic_height = None
            attempts -= 1
        if attic_height is None:
            msg = "Failed to sample valid attic height (must be less than 2.5x f2f height)."
            raise ValueError(msg)
        return attic_height

    @property
    def n_conditioned_floors(self) -> int:
        """The number of conditioned floors in the building."""
        n_floors = self.num_floors
        if self.basement_is_conditioned:
            n_floors += 1
        if self.attic_is_conditioned:
            n_floors += 1
        return n_floors

    @property
    def n_occupied_floors(self) -> int:
        """The number of occupied floors in the building."""
        n_floors = self.num_floors
        if self.basement_is_occupied:
            n_floors += 1
        if self.attic_is_occupied:
            n_floors += 1
        return n_floors

    @property
    def energy_model_footprint_area(self) -> float:
        """The floor area of the building."""
        return self.long_edge * self.short_edge

    @property
    def energy_model_conditioned_area(self) -> float:
        """The conditioned area of the building."""
        return self.n_conditioned_floors * self.energy_model_footprint_area

    @property
    def energy_model_occupied_area(self) -> float:
        """The conditioned area of the building."""
        return self.n_occupied_floors * self.energy_model_footprint_area


class GloBIOutputSpec(ExperimentOutputSpec):
    """Output for the building builder experiment."""

    hourly_data: FileReference | None = None



"""Utilities for working with weather file data."""

import logging
from collections.abc import Callable
from pathlib import Path
from typing import cast

import geopandas as gpd
import numpy as np
from scipy.spatial import cKDTree

logger = logging.getLogger(__name__)

cached_epw_metadata_singleton: gpd.GeoDataFrame | None = None


def closest_epw(
    query_pts: gpd.GeoSeries,
    source_filter: str | None = None,
    crs: str | int = 3857,
    distance_threshold_meters: int | None = 500_000,
    metadata: gpd.GeoDataFrame | None = None,
    log_fn: Callable | None = None,
) -> gpd.GeoDataFrame:
    """Find the closest epw for each row in a dataframe."""
    # Get the closest epw for each row
    if metadata is None:
        global cached_epw_metadata_singleton
        if cached_epw_metadata_singleton is None:
            metadta_path = Path(__file__).parent / "epw_metadata.geojson"
            metadata = gpd.read_file(metadta_path)
            # metadta_path = Path(__file__).parent / "epw_metadata.parquet"
            # metadata = gpd.read_parquet(metadta_path)
            cached_epw_metadata_singleton = metadata
        else:
            metadata = cached_epw_metadata_singleton

    if metadata is None:
        msg = "Failed to load metadata"
        raise ValueError(msg)

    log = log_fn or logger.info
    log(f"Querying {len(query_pts)} points for the closest EPW.")
    filtered_metadata = (
        cast(gpd.GeoDataFrame, metadata.query(source_filter))
        if source_filter
        else metadata
    )

    # Get the closest epw for each row
    filtered_metadata_projected = cast(
        gpd.GeoDataFrame,
        filtered_metadata.to_crs(crs)
        if filtered_metadata.crs != crs
        else filtered_metadata,
    )
    points: gpd.GeoSeries = filtered_metadata_projected.geometry
    is_na = points.apply(
        lambda x: np.isinf(x.x) or np.isinf(x.y) or np.isnan(x.x) or np.isnan(x.y)
    )
    points = cast(gpd.GeoSeries, points[~is_na])
    points_tuples = points.apply(lambda x: (x.x, x.y)).tolist()
    filtered_metadata = filtered_metadata[~is_na]

    query_pts_projected = query_pts.to_crs(crs) if query_pts.crs != crs else query_pts
    query_points_proj = query_pts_projected.geometry
    query_points_tuples = query_points_proj.apply(lambda x: (x.x, x.y)).tolist()

    tree = cKDTree(points_tuples)
    distance, idx = tree.query(query_points_tuples)
    distance: np.ndarray

    # Return the closest epw
    selected_metadata: gpd.GeoDataFrame = filtered_metadata.iloc[idx]
    selected_metadata = cast(gpd.GeoDataFrame, selected_metadata.copy(deep=True))
    selected_metadata["distance"] = distance
    if distance_threshold_meters:
        mask = distance < distance_threshold_meters
        if not mask.all():
            log(f"Found {len(mask) - mask.sum()} points that are too far.")
            raise ValueError("EPW:TOO_FAR")
    selected_metadata = cast(
        gpd.GeoDataFrame, selected_metadata.set_index(query_pts.index)
    )
    log(f"Found {len(selected_metadata)} points.")
    return selected_metadata



"""GloBI CLI."""

import tempfile
from pathlib import Path
from typing import TYPE_CHECKING

import boto3
import click
import yaml
from scythe.experiments import BaseExperiment, SemVer
from scythe.settings import ScytheStorageSettings

from globi.allocate import allocate_globi_dryrun
from globi.pipelines import simulate_globi_building

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client
else:
    S3Client = object


@click.group()
def cli():
    """The GloBI CLI.

    Use this to create, manage, and submit GloBI experiments.
    """
    pass


@cli.group()
def submit():
    """Submit a GloBI experiment from different sources."""
    pass


@submit.command()
@click.option(
    "--path",
    type=click.Path(exists=True),
    help="The path to the manifest file which will be used to configure the experiment.",
    prompt="Manifest file path (.yml)",
)
@click.option(
    "--scenario",
    type=str,
    help="The scenario to use for the experiment.",
    required=False,
)
@click.option(
    "--skip-model-constructability-check",
    is_flag=True,
    help="Skip the model constructability check.",
    required=False,
)
@click.option(
    "--grid-run",
    is_flag=True,
    help="Dry run the experiment allocation by only testing semantic field combinations.",
)
@click.option(
    "--epwzip-file",
    type=click.Path(exists=True),
    help="The path to the EPWZip file to use for the experiment.",
    required=False,
)
def manifest(
    path: Path,
    scenario: str | None = None,
    skip_model_constructability_check: bool = False,
    grid_run: bool = False,
    epwzip_file: Path | None = None,
):
    """Submit a GloBI experiment from a manifest file."""
    import logging

    from globi.allocate import allocate_globi_experiment
    from globi.models import GloBIExperimentSpec

    logging.basicConfig(level=logging.INFO)

    with open(path) as f:
        manifest = yaml.safe_load(f)

    config = GloBIExperimentSpec.model_validate(manifest)

    if scenario:
        config.scenario = scenario

    if epwzip_file:
        config.file_config.epwzip_file = epwzip_file

    if grid_run:
        allocate_globi_dryrun(config)
    else:
        allocate_globi_experiment(config, not skip_model_constructability_check)


@submit.command()
def artifacts():
    """Submit a GloBI experiment from a set of artifacts."""
    print("NOT IMPLEMENTED YET")
    pass


@cli.command()
@click.option(
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    help="The path to the manifest file which will be used to configure the building.",
    prompt="Manifest file path (.yml | .yaml)",
    required=True,
)
@click.option(
    "--output-dir",
    type=click.Path(file_okay=False),
    required=False,
    help="The path to the directory to use for the simulation.",
    # prompt="Output directory path (optional)",
)
def simulate(config: Path, output_dir: str | None = None):
    """Simulate a GloBI building."""
    from globi.models import GloBIBuildingSpec
    from globi.pipelines import simulate_globi_building_pipeline

    with open(config) as f:
        manifest = yaml.safe_load(f)
    conf = GloBIBuildingSpec.model_validate(manifest)
    if output_dir is None:
        print("No output directory provided, results will not be saved.")
    with tempfile.TemporaryDirectory() as tempdir:
        odir = Path(output_dir or tempdir)
        odir.mkdir(parents=True, exist_ok=True)
        epodir = odir / "ep"
        epodir.mkdir(parents=True, exist_ok=True)
        rodir = odir / "results"
        rodir.mkdir(parents=True, exist_ok=True)
        r = simulate_globi_building_pipeline(conf, epodir)
        for k, v in r.dataframes.items():
            v.to_parquet(rodir / f"{k}.parquet")
            v.reset_index(drop=True).stack(
                level="Month", future_stack=True
            ).reset_index(level=0, drop=True).to_csv(rodir / f"{k}.csv")
    if not output_dir:
        # TODO: improve results summarization
        print("Results:")
        print(
            r.dataframes["Results"]
            .Energy.T.groupby(level=["Aggregation", "Meter"])
            .sum()
            .T.reset_index(drop=True)
            .T
        )


@cli.group()
def get():
    """Get a GloBI experiment from different sources."""
    pass


@get.command()
@click.option(
    "--run-name",
    type=str,
    help="The name of the run to get.",
    required=True,
    prompt="Run name",
)
@click.option(
    "--version",
    type=str,
    help="The version of the run to get.",
    required=False,
)
@click.option(
    "--dataframe-key",
    default="Results",
    type=str,
    help="The dataframe to get.",
    required=False,
)
@click.option(
    "--output-dir",
    default="outputs",
    type=click.Path(file_okay=False),
    required=False,
    help="The path to the directory to use for the simulation.",
)
def experiment(
    run_name: str,
    version: str | None = None,
    dataframe_key: str = "Results",
    output_dir: str = "outputs",
):
    """Get a GloBI experiment from a manifest file."""
    s3_client: S3Client = boto3.client("s3")
    s3_settings = ScytheStorageSettings()
    exp = BaseExperiment(experiment=simulate_globi_building, run_name=run_name)

    if not version:
        exp_version = exp.latest_version(s3_client, from_cache=False)
        if exp_version is None:
            msg = f"No version found for experiment {run_name}"
            raise ValueError(msg)
        sem_version = exp_version.version
    else:
        sem_version = SemVer.FromString(version)

    results_filekeys = exp.latest_results_for_version(sem_version)

    if dataframe_key not in results_filekeys:
        msg = f"Dataframe key {dataframe_key} not found in results."
        raise ValueError(msg)

    output_key = Path(output_dir) / run_name / str(sem_version) / f"{dataframe_key}.pq"

    output_key.parent.mkdir(parents=True, exist_ok=True)

    s3_client.download_file(
        Bucket=s3_settings.BUCKET,
        Key=results_filekeys[dataframe_key],
        Filename=output_key.as_posix(),
    )
    print(f"Downloaded to {output_key.as_posix()}")



"""Initialize the local AWS environment."""

import logging
import os

import boto3
from scythe.settings import ScytheStorageSettings

s3 = boto3.client("s3")

logger = logging.getLogger()
logging.basicConfig(level=logging.INFO)
logger.setLevel(logging.INFO)

# list buckets
if __name__ == "__main__":
    bucket_name = ScytheStorageSettings().BUCKET
    # create the bucket if it does not exist
    try:
        region = os.getenv("AWS_DEFAULT_REGION")
        s3.create_bucket(
            Bucket=bucket_name,
            CreateBucketConfiguration={
                "LocationConstraint": region,  # pyright: ignore [reportArgumentType]
            },
        )
        logger.info(f"Bucket {bucket_name} created")
    except Exception as e:
        logger.info(f"Bucket {bucket_name} already exists, {e}")



"""Allocate a GloBI experiment with Scythe."""

import json
import logging
import math
from pathlib import Path

import boto3
import geopandas as gpd
import numpy as np
import yaml
from epinterface.sbem.fields.spec import SemanticModelFields
from epinterface.sbem.utils import check_model_existence
from scythe.base import ExperimentInputSpec, ExperimentOutputSpec
from scythe.experiments import BaseExperiment
from scythe.scatter_gather import RecursionMap
from shapely import to_wkt
from tqdm import tqdm

from globi.models import GloBIBuildingSpec, GloBIExperimentSpec
from globi.pipelines import preprocess_gis_file, simulate_globi_building

# TODO: TEST THIS!!


s3_client = boto3.client("s3")

logger = logging.getLogger(__name__)

# TODO: replace prints with logs


def allocate_globi_experiment(
    config: GloBIExperimentSpec, check_model_constructability: bool = True
):
    """Deploy an experiment from a config."""
    print("Deploying experiment from config:")
    print(yaml.dump(config.model_dump(mode="json"), indent=2, sort_keys=False))
    name = f"{config.name}/{config.scenario}"

    if check_model_constructability:
        check_model_existence(
            component_map_path=config.file_config.component_map_file,
            semantic_fields_path=config.file_config.semantic_fields_file,
            db_path=config.file_config.db_file,
            raise_on_error=True,
        )

    # if config.gis_preprocessor_config:
    buildings_gdf, colmap = preprocess_gis_file(
        # TODO: make this required on the original model
        config=config.gis_preprocessor_config,
        file_config=config.file_config,
        scenario=config.scenario,
    )

    specs: list[GloBIBuildingSpec] = []

    for sort_index, (_, row) in tqdm(
        enumerate(buildings_gdf.iterrows()),
        total=len(buildings_gdf),
        desc="Generating building specs from GIS:",
    ):
        row = row.to_dict()
        globi_spec = GloBIBuildingSpec(
            experiment_id="placeholder",
            sort_index=sort_index,
            db_file=row[colmap.DB_File_col],
            semantic_fields_file=config.file_config.semantic_fields_file,
            component_map_file=config.file_config.component_map_file,
            epwzip_file=row[colmap.EPWZip_File_col],
            semantic_field_context=row[colmap.Semantic_Field_Context_col],
            neighbor_polys=[to_wkt(poly) for poly in row[colmap.Neighbor_Polys_col]],
            neighbor_heights=row[colmap.Neighbor_Heights_col],
            neighbor_floors=row[colmap.Neighbor_Floors_col],
            rotated_rectangle=to_wkt(row[colmap.Rotated_Rectangle_col]),
            long_edge_angle=row[colmap.Long_Edge_Angle_col],
            long_edge=row[colmap.Long_Edge_col],
            short_edge=row[colmap.Short_Edge_col],
            aspect_ratio=row[colmap.Aspect_Ratio_col],
            rotated_rectangle_area_ratio=row[colmap.Rotated_Rectangle_Area_Ratio_col],
            wwr=row[colmap.WWR_col],
            height=row[colmap.Height_col],
            num_floors=row[colmap.Num_Floors_col],
            f2f_height=row[colmap.F2F_Height_col],
            basement=row[colmap.Basement_col],
            attic=row[colmap.Attic_col],
            exposed_basement_frac=row[colmap.Exposed_Basement_Frac_col],
            parent_experiment_spec=config,
        )
        specs.append(globi_spec)

    if not specs:
        msg = "No specs provided"
        raise ValueError(msg)

    experiment = BaseExperiment[ExperimentInputSpec, ExperimentOutputSpec](
        experiment=simulate_globi_building, run_name=name
    )
    print(f"Submitting {len(buildings_gdf)} buildings for experiment {name}")
    min_branches_required, _, _ = calculate_branching_factor(specs)

    run, ref = experiment.allocate(
        specs,
        version="bumpminor",
        recursion_map=RecursionMap(factor=min_branches_required, max_depth=1),
        s3_client=s3_client,
    )

    print(yaml.dump(run.model_dump(mode="json"), indent=2, sort_keys=False))
    return run, ref


def allocate_globi_dryrun(
    config: GloBIExperimentSpec, epwzip_file: Path | str | None = None
):
    """Dry run the allocation of an experiment to estimate the cost."""
    from shapely import Polygon, to_wkt

    epwzip_file = epwzip_file or config.file_config.epwzip_file
    if epwzip_file is None:
        msg = "EPWZip file is required for dry run"
        raise ValueError(msg)

    max_tests = 1000
    with open(config.file_config.semantic_fields_file) as f:
        model = SemanticModelFields.model_validate(yaml.safe_load(f))

    grid, field_vals = model.make_grid(numerical_discretization=10)
    grid = grid.sample(min(max_tests, len(grid)))

    width = 12  # meters
    basic_rectangle = Polygon([(0, 0), (width, 0), (width, width), (0, width), (0, 0)])

    specs: list[GloBIBuildingSpec] = []
    for _, row in grid.iterrows():
        context = row.to_dict()
        # first we need to update the context with the field values
        for field_name, field_val in field_vals.items():
            context[field_name] = field_val[context[field_name]]
        spec = GloBIBuildingSpec(
            experiment_id="placeholder",
            sort_index=0,
            db_file=config.file_config.db_file,
            semantic_fields_file=config.file_config.semantic_fields_file,
            component_map_file=config.file_config.component_map_file,
            epwzip_file=epwzip_file,  # pyright: ignore [reportArgumentType]
            semantic_field_context=context,
            neighbor_polys=[],
            neighbor_heights=[],
            neighbor_floors=[],
            rotated_rectangle=to_wkt(basic_rectangle),
            long_edge_angle=0,
            long_edge=width,
            short_edge=width,
            aspect_ratio=1,
            rotated_rectangle_area_ratio=width * width,
            wwr=config.gis_preprocessor_config.default_wwr,
            height=config.gis_preprocessor_config.default_num_floors
            * config.gis_preprocessor_config.f2f_height,
            num_floors=config.gis_preprocessor_config.default_num_floors,
            f2f_height=config.gis_preprocessor_config.f2f_height,
            basement=config.gis_preprocessor_config.default_basement,
            attic=config.gis_preprocessor_config.default_attic,
            exposed_basement_frac=config.gis_preprocessor_config.default_exposed_basement_frac,
        )
        specs.append(spec)

    min_branches_required, _, _ = calculate_branching_factor(specs)

    if not specs:
        msg = "No specs provided"
        raise ValueError(msg)

    experiment = BaseExperiment[ExperimentInputSpec, ExperimentOutputSpec](
        experiment=simulate_globi_building,
        run_name=f"{config.name}/dryrun/{config.scenario}",
    )

    run, ref = experiment.allocate(
        specs,
        version="bumpminor",
        recursion_map=RecursionMap(factor=min_branches_required, max_depth=1),
        s3_client=s3_client,
    )

    print(yaml.dump(run.model_dump(mode="json"), indent=2, sort_keys=False))
    return run, ref


def calculate_branching_factor(specs: list[GloBIBuildingSpec]) -> tuple[int, int, int]:
    """Calculate the branching factor for a given list of building specs.

    We do this by sampling 1k random buildings and checking the size of their serialized payloads.

    This is necessary because the async fanouts send all of the payloads for a branch over the wire at once.
    """
    logger.info("Calculating branching factor...")
    ixs = np.random.choice(len(specs), size=1000, replace=True)
    total_bytes = 0
    for ix in ixs:
        # check the file size of json.sumps
        stringified = json.dumps(specs[ix].model_dump(mode="json"), indent=2)
        nbytes = len(stringified.encode("utf-8"))
        total_bytes += nbytes
    avg_bytes = total_bytes / len(ixs)
    max_bytes_MB = 3  # safety factor, 4MB is actual amx
    max_bytes_B = max_bytes_MB * 1024 * 1024
    sims_per_branch = math.floor(max_bytes_B / avg_bytes)
    min_branches_required = math.ceil(len(specs) / sims_per_branch)
    logger.info(f"Avg payload size: {int(avg_bytes // 1024):0d} kB")
    logger.info(f"Avg sims per branch: {sims_per_branch}")
    logger.info(f"Min branches required: {min_branches_required}")
    return min_branches_required, sims_per_branch, math.ceil(avg_bytes)


if __name__ == "__main__":
    import logging

    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)
    spec = GloBIExperimentSpec.from_(Path("data/partners/Cambridge_UK/manifest.yml"))
    print(yaml.dump(spec.model_dump(mode="json"), indent=2, sort_keys=False))
    # gdf = gpd.read_file(spec.file_config.gis_file)
    logging.basicConfig(level=logging.DEBUG)
    if spec.gis_preprocessor_config is not None:
        new_gdf_path = spec.file_config.gis_file.parent / "buildings-updated.geojson"
        old_gdf: gpd.GeoDataFrame = gpd.read_file(spec.file_config.gis_file)
        old_gdf["Region"] = "CB"
        old_gdf["Age_bracket"] = (
            old_gdf["Age_bracket"].str.replace(" ", "_").str.replace("1999", "1960")
        )
        old_gdf["OCCUPANCY_DENSITY"] = old_gdf["OCCUPANCY_DENSITY"].str.replace(
            "MediumDensity", "MedDensity"
        )
        old_gdf.to_file(new_gdf_path, driver="GeoJSON")
        spec.file_config.gis_file = new_gdf_path
        allocate_globi_experiment(spec)



"""Utilities for GIS preprocessor."""

import logging
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import cast, get_args
from urllib.parse import urljoin

import geopandas as gpd
import numpy as np
import pandas as pd
from epinterface.sbem.fields.spec import (
    CategoricalFieldSpec,
    NumericFieldSpec,
    SemanticModelFields,
)

from globi.gis.errors import (
    GISFileHasInvalidCategoricalSemanticFieldError,
    GISFileHasMissingBuildingIDsError,
    GISFileHasMissingColumnsError,
    GISFileHasNoBuildingIDColumnError,
    GISFileHasNoCRSError,
    GISFileHasNoFeaturesError,
    GISFileHasNonNumericFloorsError,
    GISFileHasNonNumericHeightError,
    GISFileHasNonUniqueBuildingIDsError,
    GISFileHasOutOfBoundsNumericSemanticFieldError,
    GISFileHasSemanticFieldWithNoValidatorError,
    GISFileHasUnexpectedCRSError,
    GISFileMissingBothHeightAndFloorsError,
)
from globi.gis.weather import closest_epw
from globi.type_utils import BasementAtticOccupationConditioningStatus

logger = logging.getLogger(__name__)


LogLikeFn = Callable[[str], None]


def reproject_gdf(
    gdf: gpd.GeoDataFrame, cart_crs: str, log_fn: LogLikeFn | None = None
) -> gpd.GeoDataFrame:
    """We want to have some safety to ensure there is no confusing behavior with the provided CRS.

    We want to ensure that either:
    - the gdf is already in the desired crs or EPSG:3857, in which case we bring it back up to 4326
    - the gdf is already in 4326, in which case we leave it alone
    - the gdf has no crs, in which case we raise an error
    - the gdf is in some other crs, in which case we raise an error

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        cart_crs (str): The eventually desired cartesian crs.
        log_fn (Callable | None): The function to use for logging.

    Raises:
        GISFileHasNoCRSError: If the gdf has no crs.
        GISFileHasUnexpectedCRSError: If the gdf is in some other crs.

    Returns:
        gdf (gpd.GeoDataFrame): The reprojected GeoDataFrame.

    """
    log = log_fn or logger.info
    log("Checking GIS file crs and reprojecting if necessary...")
    if not gdf.crs:
        raise GISFileHasNoCRSError()

    if gdf.crs == "EPSG:3857":
        log("Reprojecting gis file to EPSG:4326 from EPSG:3857.")
        gdf.to_crs("EPSG:4326", inplace=True)
    current_crs = gdf.crs

    log(f"GIS file has crs {current_crs}")
    if str(current_crs) not in ["EPSG:4326", str(cart_crs)]:
        raise GISFileHasUnexpectedCRSError(str(cart_crs), str(current_crs))

    if current_crs != "EPSG:4326":
        log("Projecting gis file to EPSG:4326 from {current_crs}.")
        gdf.to_crs("EPSG:4326", inplace=True)

    return gdf


def rename_shp_cols(
    gdf: gpd.GeoDataFrame,
    expected_cols: list[str | None],
    log_fn: LogLikeFn | None = None,
) -> gpd.GeoDataFrame:
    """Rename columns of a shapefile to the expected column names.

    This is necessary because shapefiles will truncate the column
    name to 10 characters, but users might not realize this when they
    export from e.g. ArcGIS.

    """
    log = log_fn or logger.info
    log("Renaming SHAPEFILE columns which may have been truncated...")
    for col_name in expected_cols:
        if col_name is None:
            continue
        if col_name[:10] in gdf.columns:
            if col_name[:10] != col_name:
                log(
                    f"Renaming column '{col_name[:10]}' to '{col_name}' as per semantic fields."
                )
            gdf.rename(columns={col_name[:10]: col_name}, inplace=True)
    log("Done renaming SHAPEFILE columns.")
    return gdf


def validate_semantic_field_compatibility(
    gdf: gpd.GeoDataFrame,
    semantic_fields_data: SemanticModelFields,
    missing_ok: bool = True,
    log_fn: LogLikeFn | None = None,
) -> None:
    """Validate the user provided semantic fields against the provided GeoDataFrame.

    Args:
        gdf: The input GeoDataFrame.
        semantic_fields_data: The semantic fields data loaded from YAML.
        missing_ok: Whether to allow missing values for a field.
        log_fn: The function to use for logging.

    Raises:
        GISFileHasInvalidCategoricalSemanticFieldError: If an option is not found in the allowed options for a given semantic field.
        GISFileHasOutOfBoundsNumericSemanticFieldError: If a numeric value is out of bounds.
        GISFileHasSemanticFieldWithNoValidatorError: If a semantic field has no validator.
    """
    log = log_fn or logger.info
    for field in semantic_fields_data.Fields:
        field_name = field.Name
        if field_name not in gdf.columns:
            continue
        if isinstance(field, CategoricalFieldSpec):
            valid_opts = field.Options
            if valid_opts:
                field_vals = gdf[field_name]
                is_missing = field_vals.isna()

                is_valid = cast(
                    pd.Series,
                    field_vals.isin(valid_opts)
                    | (is_missing & pd.Series([missing_ok] * len(field_vals))),
                )
                is_not_valid = ~is_valid
                if cast(pd.Series, is_not_valid).any():
                    invalid_series = cast(pd.Series, field_vals[is_not_valid])
                    invalid_values = invalid_series.unique().tolist()
                    raise GISFileHasInvalidCategoricalSemanticFieldError(
                        field_name, invalid_values, valid_opts
                    )
        elif isinstance(field, NumericFieldSpec):
            min_val: float = field.Min
            max_val: float = field.Max
            if min_val is not None and max_val is not None:
                field_vals = cast(pd.Series, gdf[field_name])
                is_missing = field_vals.isna()
                is_valid = cast(
                    pd.Series,
                    field_vals.between(min_val, max_val, inclusive="both")
                    | (is_missing & pd.Series([missing_ok] * len(field_vals))),
                )
                is_not_valid = ~is_valid
                if cast(pd.Series, is_not_valid).any():
                    invalid_series = cast(pd.Series, field_vals[is_not_valid])
                    invalid_values = invalid_series.unique().tolist()
                    raise GISFileHasOutOfBoundsNumericSemanticFieldError(
                        field_name, invalid_values, min_val, max_val
                    )
        else:
            raise GISFileHasSemanticFieldWithNoValidatorError(field_name)

    log("Semantic field compatibility validation complete.")


def validate_has_rows(gdf: gpd.GeoDataFrame) -> None:
    """Validate that the GeoDataFrame has rows."""
    if len(gdf) == 0:
        raise GISFileHasNoFeaturesError()


def check_for_column_existence(
    gdf: gpd.GeoDataFrame,
    expected_cols: Sequence[str | None],
    log_fn: LogLikeFn | None = None,
) -> None:
    """Check if the expected columns exist in the GeoDataFrame.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        expected_cols (Sequence[str | None]): The expected column names.
        log_fn (Callable | None): The function to use for logging.

    Raises:
        GISFileHasMissingColumnsError: If a column is not found in the GeoDataFrame.
    """
    log = log_fn or logger.info
    log("Checking for column existence...")
    missing_cols = []
    for col_name in expected_cols:
        if col_name is None:
            continue
        if col_name not in gdf.columns:
            missing_cols.append(col_name)
        else:
            log(f"Column '{col_name}' found in gdf.")
    if missing_cols:
        raise GISFileHasMissingColumnsError(missing_cols, gdf.columns.tolist())
    log("All expected columns found in gdf.")


def check_building_ids(
    gdf: gpd.GeoDataFrame,
    building_id_col: str,
    log_fn: LogLikeFn | None = None,
) -> tuple[gpd.GeoDataFrame, str]:
    """Inject building ids into the GeoDataFrame if they are not already present.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        building_id_col (str): The name of the building id column.
        log_fn (LogLikeFn | None): The function to use for logging.

    Raises:
        GISFileHasNoBuildingIDColumnError: If the building id column is not found.
        GISFileHasMissingBuildingIDsError: If the building id column has missing values.
        GISFileHasNonUniqueBuildingIDsError: If the building id column has non-unique values.

    Returns:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame with the building ids.
        building_id_col (str): The name of the building id column.
    """
    log = log_fn or logger.info
    log("Checking for building ids...")

    # If the building id column is not present, we will inject uuids
    if building_id_col not in gdf.columns:
        log("Building id column not found.")
        raise GISFileHasNoBuildingIDColumnError()

    # If there are missing ids, we will back fill them with new uuids
    has_missing_ids = cast(pd.Series, gdf[building_id_col].isna())
    if has_missing_ids.any():
        raise GISFileHasMissingBuildingIDsError()

    gdf[building_id_col] = gdf[building_id_col].astype(str)
    if gdf[building_id_col].nunique() != len(gdf):
        raise GISFileHasNonUniqueBuildingIDsError()

    return gdf, building_id_col


def drop_by_area(
    gdf: gpd.GeoDataFrame,
    area_col: str,
    min_area: float,
    log_fn: LogLikeFn | None = None,
) -> tuple[gpd.GeoDataFrame, int]:
    """Drop features with an area less than the minimum area.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        area_col (str): The name of the area column.
        min_area (float): The minimum area.
        log_fn (LogLikeFn | None): The function to use for logging.

    Returns:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame with the dropped features.
        n_dropped (int): The number of features dropped.
    """
    log = log_fn or logger.info
    log("Checking that features have an area greater than the minimum area...")
    is_valid_area = cast(pd.Series, gdf[area_col] >= min_area)
    n_dropped = len(gdf) - is_valid_area.sum()
    gdf = cast(gpd.GeoDataFrame, gdf[is_valid_area])
    if n_dropped > 0:
        log(
            f"Dropped {n_dropped / len(gdf):.1%} of remaining features with an area less than {min_area} (n={n_dropped})."
        )
    else:
        log("All features have an area greater than the minimum area.")
    return gdf, n_dropped


def drop_by_edge_length(
    gdf: gpd.GeoDataFrame,
    min_edge_length_m: float,
    max_edge_length_m: float,
    short_edge_col: str,
    long_edge_col: str,
    log_fn: LogLikeFn | None = None,
) -> tuple[gpd.GeoDataFrame, int]:
    """Drop features with an edge length less than the minimum or greater than the maximum edge length.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        min_edge_length_m (float): The minimum edge length.
        max_edge_length_m (float): The maximum edge length.
        short_edge_col (str): The name of the short edge column.
        long_edge_col (str): The name of the long edge column.
        log_fn (LogLikeFn | None): The function to use for logging.

    Returns:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame with the dropped features.
        n_dropped (int): The number of features dropped.
    """
    log = log_fn or logger.info
    log(
        "Checking that features have an edge length between the minimum and maximum edge length..."
    )
    is_valid_short_edge = cast(pd.Series, gdf[short_edge_col]).between(
        min_edge_length_m, max_edge_length_m, inclusive="both"
    )
    is_valid_long_edge = cast(pd.Series, gdf[long_edge_col]).between(
        min_edge_length_m, max_edge_length_m, inclusive="both"
    )
    is_valid_edge = is_valid_short_edge & is_valid_long_edge
    n_dropped = len(gdf) - is_valid_edge.sum()
    gdf = cast(gpd.GeoDataFrame, gdf[is_valid_edge])
    if n_dropped > 0:
        log(
            f"Dropped {n_dropped / len(gdf):.1%} of remaining features with an edge length less than {min_edge_length_m} or greater than {max_edge_length_m} (n={n_dropped})."
        )
    else:
        log(
            "All features have an edge length between the minimum and maximum edge length."
        )
    return gdf, n_dropped


def add_lat_and_lon_cols(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Add latitude and longitude columns to the GeoDataFrame."""
    # TODO: do we care that we may be overwriting here?
    centroids = cast(gpd.GeoDataFrame, gdf.to_crs("EPSG:3857")).geometry.centroid
    points = gpd.GeoSeries(centroids)
    points.crs = "EPSG:3857"
    points = points.to_crs("EPSG:4326")
    gdf["lat"] = points.y
    gdf["lon"] = points.x
    return gdf


def handle_height_and_floors(
    gdf: gpd.GeoDataFrame,
    height_col: str | None,
    nfloors_col: str | None,
    assumed_f2f_height: float,
    default_n_floors: int,
    min_floors: int,
    max_floors: int,
    min_height: float,
    max_height: float,
    log_fn: LogLikeFn | None = None,
) -> tuple[gpd.GeoDataFrame, str, str, str, int]:
    """Handle the height and floors columns.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        height_col (str | None): The name of the height column.
        nfloors_col (str | None): The name of the floors column.
        assumed_f2f_height (float): The assumed height of a floor to ceiling height.
        default_n_floors (int): The default number of floors.
        min_floors (int): The minimum number of floors.
        max_floors (int): The maximum number of floors.
        min_height (float): The minimum height.
        max_height (float): The maximum height.
        log_fn (LogLikeFn | None): The function to use for logging.

    Returns:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame with the height and floors columns.
        height_col (str): The name of the height column.
        nfloors_col (str): The name of the floors column.
        n_dropped (int): The number of buildings dropped due to invalid floors.
    """
    log = log_fn or logger.info

    log("Checking that height and floors columns have data...")

    if height_col is None and nfloors_col is None:
        raise GISFileMissingBothHeightAndFloorsError()

    if height_col is not None and gdf[height_col].dtype not in [
        float,
        int,
        np.float64,
        np.int64,
        np.float32,
        np.int32,
    ]:
        raise GISFileHasNonNumericHeightError(height_col)

    if nfloors_col is not None and gdf[nfloors_col].dtype not in [
        float,
        int,
        np.float64,
        np.int64,
        np.float32,
        np.int32,
    ]:
        raise GISFileHasNonNumericFloorsError(nfloors_col)

    if height_col is None:
        log(
            f"No height column provided, will impute based off of number floors and assumed floor-to-floor height ({assumed_f2f_height}m)..."
        )
        log(
            f"(Buildings with unknown floor counts will be assumed to have {default_n_floors} floors.)"
        )
        nfloors_col = nfloors_col or "GLOBI_NFLOORS"
        existing_floor_data = cast(
            pd.Series, gdf[nfloors_col].round(0).astype(int)
        ).fillna(default_n_floors)
        gdf[nfloors_col] = existing_floor_data
        new_height_data = existing_floor_data * assumed_f2f_height
        height_col = "GLOBI_HEIGHT"
        gdf[height_col] = new_height_data

    elif nfloors_col is None:
        log(
            f"No floors column provided, will impute based off of building height and assumed floor-to-floor height ({assumed_f2f_height}m)..."
        )
        log(
            f"(Buildings with unknown heights will be assumed to have {default_n_floors * assumed_f2f_height}m height.)"
        )
        height_col = height_col or "GLOBI_HEIGHT"
        existing_height_data = cast(pd.Series, gdf[height_col]).fillna(
            default_n_floors * assumed_f2f_height
        )
        gdf[height_col] = existing_height_data
        nfloors_col = "GLOBI_NFLOORS"
        gdf[nfloors_col] = (gdf[height_col] / assumed_f2f_height).round(0).astype(int)
        gdf[height_col] = gdf[nfloors_col] * assumed_f2f_height

    is_valid_height = cast(pd.Series, gdf[height_col]).between(
        min_height, max_height, inclusive="both"
    )
    is_valid_floors = cast(pd.Series, gdf[nfloors_col]).between(
        min_floors, max_floors, inclusive="both"
    )
    is_valid = is_valid_height & is_valid_floors
    n_dropped = len(gdf) - is_valid.sum()
    gdf = cast(gpd.GeoDataFrame, gdf[is_valid])
    if n_dropped > 0:
        log(
            f"Dropped {n_dropped / len(gdf):.1%} of remaining features with a height or floors out of bounds (n={n_dropped})."
        )
    f2f_height_col = "GLOBI_F2F_HEIGHT"
    gdf[f2f_height_col] = assumed_f2f_height
    log("Done handling height and floors columns.")
    return gdf, height_col, nfloors_col, f2f_height_col, n_dropped


def handle_wwr(
    gdf: gpd.GeoDataFrame,
    wwr_col: str | None,
    assumed_wwr: float,
    log_fn: LogLikeFn | None = None,
) -> tuple[gpd.GeoDataFrame, str, int]:
    """Handle the window-to-wall ratio column.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        wwr_col (str | None): The name of the window-to-wall ratio column.
        assumed_wwr (float): The assumed window-to-wall ratio.
        log_fn (LogLikeFn | None): The function to use for logging.

    Returns:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame with the window-to-wall ratio column.
        wwr_col (str): The name of the window-to-wall ratio column.
        n_dropped (int): The number of buildings dropped due to invalid window-to-wall ratio.
    """
    log = log_fn or logger.info
    log("Checking that window-to-wall ratio column has data...")
    if wwr_col is None:
        log(
            f"No window-to-wall ratio column provided, will impute based off of assumed window-to-wall ratio ({assumed_wwr})..."
        )
        wwr_col = "GLOBI_WWR"
        gdf[wwr_col] = assumed_wwr
    gdf[wwr_col] = cast(pd.Series, gdf[wwr_col]).fillna(assumed_wwr)
    # TODO: add a dtype checker
    min_wwr = 0
    max_wwr = 1
    is_valid = cast(pd.Series, gdf[wwr_col]).between(min_wwr, max_wwr, inclusive="both")
    n_dropped = len(gdf) - is_valid.sum()
    gdf = cast(gpd.GeoDataFrame, gdf[is_valid])
    if n_dropped > 0:
        log(
            f"Dropped {n_dropped / len(gdf):.1%} of remaining features with a window-to-wall ratio out of bounds ({min_wwr} - {max_wwr}) (n={n_dropped})."
        )
    log("Done handling window-to-wall ratio column.")
    return gdf, wwr_col, n_dropped


def handle_basement(
    gdf: gpd.GeoDataFrame,
    basement_col: str | None,
    assumed_basement: BasementAtticOccupationConditioningStatus,
    log_fn: LogLikeFn | None = None,
) -> tuple[gpd.GeoDataFrame, str, int]:
    """Handle the basement column.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        basement_col (str | None): The name of the basement column.
        assumed_basement (BasementAtticOccupationConditioningStatus): The assumed basement status.
        log_fn (LogLikeFn | None): The function to use for logging.

    Returns:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame with the basement column.
        basement_col (str): The name of the basement column.
        n_dropped (int): The number of buildings dropped due to invalid basement status.
    """
    log = log_fn or logger.info
    log("Checking that basement column has data...")
    if basement_col is None:
        log(
            f"No basement column provided, will impute based off of assumed basement status ('{assumed_basement}')..."
        )
        basement_col = "GLOBI_BASEMENT"
        gdf[basement_col] = assumed_basement
    # TODO: add count of NA's filled
    gdf[basement_col] = cast(pd.Series, gdf[basement_col]).fillna(assumed_basement)
    is_valid = cast(
        pd.Series,
        gdf[basement_col].isin(get_args(BasementAtticOccupationConditioningStatus)),
    )
    n_dropped = len(gdf) - is_valid.sum()
    gdf = cast(gpd.GeoDataFrame, gdf[is_valid])
    if n_dropped > 0:
        log(
            f"Dropped {n_dropped / len(gdf):.1%} of remaining features with a basement status value not in {get_args(BasementAtticOccupationConditioningStatus)} (n={n_dropped})."
        )
    log("Done handling basement column.")
    return gdf, basement_col, n_dropped


def handle_basement_exposed_fraction(
    gdf: gpd.GeoDataFrame,
    basement_exposed_fraction_col: str | None,
    assumed_basement_exposed_fraction: float,
    log_fn: LogLikeFn | None = None,
) -> tuple[gpd.GeoDataFrame, str, int]:
    """Handle the basement exposed fraction column.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        basement_exposed_fraction_col (str | None): The name of the basement exposed fraction column.
        assumed_basement_exposed_fraction (float): The assumed basement exposed fraction.
        log_fn (LogLikeFn | None): The function to use for logging.

    Returns:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame with the basement exposed fraction column.
        basement_exposed_fraction_col (str): The name of the basement exposed fraction column.
        n_dropped (int): The number of buildings dropped due to invalid basement exposed fraction.
    """
    log = log_fn or logger.info
    log("Checking that basement exposed fraction column has data...")
    if basement_exposed_fraction_col is None:
        log(
            f"No basement exposed fraction column provided, will impute based off of assumed basement exposed fraction ('{assumed_basement_exposed_fraction}')..."
        )
        basement_exposed_fraction_col = "GLOBI_EXPOSED_BASEMENT_FRAC"
        gdf[basement_exposed_fraction_col] = assumed_basement_exposed_fraction
    gdf[basement_exposed_fraction_col] = cast(
        pd.Series, gdf[basement_exposed_fraction_col]
    ).fillna(assumed_basement_exposed_fraction)
    is_valid = cast(pd.Series, gdf[basement_exposed_fraction_col]).between(
        0, 1, inclusive="both"
    )
    n_dropped = len(gdf) - is_valid.sum()
    gdf = cast(gpd.GeoDataFrame, gdf[is_valid])
    if n_dropped > 0:
        log(
            f"Dropped {n_dropped / len(gdf):.1%} of remaining features with a basement exposed fraction out of bounds (0 - 1) (n={n_dropped})."
        )
    log("Done handling basement exposed fraction column.")
    return gdf, basement_exposed_fraction_col, n_dropped


def handle_attic(
    gdf: gpd.GeoDataFrame,
    attic_col: str | None,
    assumed_attic: BasementAtticOccupationConditioningStatus,
    log_fn: LogLikeFn | None = None,
) -> tuple[gpd.GeoDataFrame, str, int]:
    """Handle the attic column.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        attic_col (str | None): The name of the attic column.
        assumed_attic (BasementAtticOccupationConditioningStatus): The assumed attic status.
        log_fn (LogLikeFn | None): The function to use for logging.

    Returns:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame with the attic column.
        attic_col (str): The name of the attic column.
        n_dropped (int): The number of buildings dropped due to invalid attic status.
    """
    log = log_fn or logger.info
    log("Checking that attic column has data...")
    if attic_col is None:
        log(
            f"No attic column provided, will impute based off of assumed attic status ('{assumed_attic}')..."
        )
        attic_col = "GLOBI_ATTIC"
        gdf[attic_col] = assumed_attic
    gdf[attic_col] = cast(pd.Series, gdf[attic_col]).fillna(assumed_attic)
    is_valid = cast(
        pd.Series,
        gdf[attic_col].isin(get_args(BasementAtticOccupationConditioningStatus)),
    )
    n_dropped = len(gdf) - is_valid.sum()
    gdf = cast(gpd.GeoDataFrame, gdf[is_valid])
    if n_dropped > 0:
        log(
            f"Dropped {n_dropped / len(gdf):.1%} of remaining features with an attic status value not in {get_args(BasementAtticOccupationConditioningStatus)} (n={n_dropped})."
        )
    log("Done handling attic column.")
    return gdf, attic_col, n_dropped


def drop_non_polygons(
    gdf: gpd.GeoDataFrame,
    log_fn: LogLikeFn | None = None,
) -> tuple[gpd.GeoDataFrame, int]:
    """Drop any features which are points."""
    log = log_fn or logger.info
    log("Checking that features are not points...")
    is_valid = cast(pd.Series, cast(gpd.GeoSeries, gdf.geometry).geom_type).isin([
        "Polygon",
        "MultiPolygon",
    ])
    n_dropped = len(gdf) - is_valid.sum()
    gdf = cast(gpd.GeoDataFrame, gdf[is_valid])
    if n_dropped > 0:
        log(
            f"Dropped {n_dropped / len(gdf):.1%} of remaining features which are not polygons (n={n_dropped})."
        )
    log("Done dropping non-polygons.")
    return gdf, n_dropped


def handle_epwzip(
    gdf: gpd.GeoDataFrame,
    weather_file_col: str | None,
    assumed_epwzip: Path | str | None,
    epw_query: str | None,
    cart_crs: str,
    log_fn: LogLikeFn | None = None,
) -> tuple[gpd.GeoDataFrame, str]:
    """Handle the epwzip column.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        weather_file_col (str | None): The name of the weather file column.
        assumed_epwzip (Path | str | None): The assumed EPW file.
        epw_query (str | None): The EPW query.
        cart_crs (str): The cartographic CRS.
        log_fn (LogLikeFn | None): The function to use for logging.

    Returns:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame with the epwzip column.
        weather_file_col (str): The name of the weather file column.
    """
    log = log_fn or logger.info
    crs = gdf.crs
    if str(crs) != "EPSG:4326":
        raise GISFileHasUnexpectedCRSError(
            expected_crs="EPSG:4326", actual_crs=str(crs)
        )

    # Precedence order:
    # 1. whatever is in the existing epwzip_file column
    # 2. the assumed epwzip
    # 3. the closest epw to the centroid of the building
    weather_file_col = weather_file_col or "GLOBI_EPWZIP"
    if weather_file_col not in gdf.columns:
        gdf[weather_file_col] = assumed_epwzip
    if assumed_epwzip is not None:
        gdf[weather_file_col] = cast(pd.Series, gdf[weather_file_col]).fillna(
            assumed_epwzip  # pyright: ignore [reportCallIssue, reportArgumentType]
        )

    is_na = cast(pd.Series, gdf[weather_file_col].isna())
    if is_na.sum() > 0:
        log("resolving remaining EPW files based on building locations...")
        needs_epw_gdf = cast(gpd.GeoDataFrame, gdf[is_na])
        # TODO: avoid calling centroid, instead construct points from lat lon
        epw_meta = closest_epw(
            needs_epw_gdf.geometry.centroid,
            source_filter=epw_query,
            crs=cart_crs,
            log_fn=log,
        )

        def handle_epw_path(x: str) -> str:
            """Format EPW path to full URL."""
            # TODO: updated this from the og method for path normalization - discuss if this method will work for all edge cases
            base_url = "https://climate.onebuilding.org/"
            return urljoin(base_url, x)

        gdf.loc[is_na, weather_file_col] = epw_meta["path"].apply(handle_epw_path)
    return gdf, weather_file_col


def inject_semantic_fields(
    gdf: gpd.GeoDataFrame,
    semantic_fields: SemanticModelFields,
) -> tuple[gpd.GeoDataFrame, str]:
    """Inject the semantic fields into the GeoDataFrame.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.
        semantic_fields (SemanticModelFields): The semantic fields.

    Returns:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame with the semantic fields.
        semantic_fields_file_col (str): The name of the semantic fields file column.
    """
    semantic_fields_context_col = "GLOBI_SEMANTIC_FIELDS_CONTEXT"
    gdf[semantic_fields_context_col] = gdf.apply(
        lambda row: {
            field_name: row[field_name]
            for field_name in semantic_fields.semantic_field_names
        },
        axis=1,
    )
    return gdf, semantic_fields_context_col
